# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from typing import TextIO, Callable

from trashcli.lib.my_input import Input
from trashcli.lib.print_version import PrintVersionAction, PrintVersionArgs
from trashcli.restore.args import RunRestoreArgs
from trashcli.restore.file_system import RestoreReadFileSystem, \
    RestoreWriteFileSystem, ReadCwd
from trashcli.restore.handler import HandlerImpl
from trashcli.restore.real_output import RealOutput
from trashcli.restore.restore_arg_parser import RestoreArgParser
from trashcli.restore.restorer import Restorer
from trashcli.restore.run_restore_action import RunRestoreAction, Handler
from trashcli.restore.trashed_files import TrashedFiles


class RestoreCmd(object):
    @staticmethod
    def make(stdout,  # type: TextIO
             stderr,  # type: TextIO
             exit,  # type: Callable[[int], None]
             input,  # type: Input
             version,  # type: str
             trashed_files,  # type: TrashedFiles
             read_fs,  # type: RestoreReadFileSystem
             write_fs,  # type: RestoreWriteFileSystem
             read_cwd,  # type: ReadCwd
             ):  # type: (...) -> RestoreCmd
        restorer = Restorer(read_fs, write_fs)
        output = RealOutput(stdout, stderr, exit)
        handler = HandlerImpl(input, read_cwd, restorer, output)
        return RestoreCmd(stdout, version, trashed_files, read_cwd, handler)

    def __init__(self,
                 stdout,  # type: TextIO
                 version,  # type: str
                 trashed_files,  # type: TrashedFiles
                 read_cwd,  # type: ReadCwd
                 handler,  # type: Handler
                 ):
        self.read_cwd = read_cwd
        self.parser = RestoreArgParser()
        self.run_restore_action = RunRestoreAction(handler,
                                                   trashed_files)
        self.print_version_action = PrintVersionAction(stdout,
                                                       version)

    def run(self, argv):
        args = self.parser.parse_restore_args(argv,
                                              self.read_cwd.getcwd_as_realpath())

        if isinstance(args, RunRestoreArgs):
            self.run_restore_action.run_action(args)
        elif isinstance(args, PrintVersionArgs):
            self.print_version_action.run_action(args)



import os

from trashcli.restore.file_system import ListingFileSystem


class InfoFiles:
    def __init__(self,
                 fs,  # type: ListingFileSystem
                 ):
        self.fs = fs

    def all_info_files(self, path):
        norm_path = os.path.normpath(path)
        info_dir = os.path.join(norm_path, 'info')
        try:
            for info_file in self.fs.list_files_in_dir(info_dir):
                if not os.path.basename(info_file).endswith('.trashinfo'):
                    yield ('non_trashinfo', info_file)
                else:
                    yield ('trashinfo', info_file)
        except OSError:  # when directory does not exist
            pass



from trashcli.lib.my_input import Input
from trashcli.put.describer import Describer


class User:
    def __init__(self,
                 my_input,  # type: Input
                 describer,  # type: Describer
                 ):
        self.input = my_input
        self.describer = describer


    def ask_user_about_deleting_file(self, program_name, path):
        reply = self.input.read_input(
            "%s: trash %s '%s'? " % (program_name,
                                     self.describer.describe(path), path))
        return parse_user_reply(reply)


user_replied_no = "user_replied_no"
user_replied_yes = "user_replied_yes"


def parse_user_reply(reply):
    return {False: user_replied_no,
            True: user_replied_yes}[reply.lower().startswith("y")]



from typing import NamedTuple, List

from trashcli.restore.index import Sequence


class Sequences(NamedTuple('Sequences', [
    ('sequences', List[Sequence]),
])):
    def all_indexes(self):
        for sequence in self.sequences:
            for index in sequence:
                yield index



from typing import NamedTuple, TypeVar

from trashcli.put.suffix import Suffix

from trashcli.put.core.int_generator import IntGenerator

from trashcli.put.my_logger import LoggerBackend

from trashcli.put.clock import PutClock

from trashcli.lib.environ import Environ
from trashcli.put.core.candidate import Candidate
from trashcli.put.core.either import Left
from trashcli.put.core.failure_reason import FailureReason, LogContext
from trashcli.put.core.trashee import Trashee
from trashcli.put.fs.fs import Fs
from trashcli.put.janitor_tools.info_creator import TrashInfoCreator
from trashcli.put.janitor_tools.info_file_persister import InfoFilePersister, \
    TrashedFile
from trashcli.put.janitor_tools.put_trash_dir import PutTrashDir
from trashcli.put.janitor_tools.security_check import SecurityCheck
from trashcli.put.janitor_tools.trash_dir_checker import TrashDirChecker
from trashcli.put.janitor_tools.trash_dir_creator import TrashDirCreator
from trashcli.put.jobs import JobExecutor
from trashcli.put.my_logger import LogData
from trashcli.put.my_logger import MyLogger


class NoLog(FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return ""


class Janitor:
    def __init__(self,
                 fs,  # type: Fs
                 clock,  # type: PutClock
                 backend,  # type: LoggerBackend
                 randint,  # type: IntGenerator
                 ):
        self.trash_dir = PutTrashDir(fs)
        self.trashing_checker = TrashDirChecker(fs)
        self.info_dir = TrashInfoCreator(fs, clock)
        self.security_check = SecurityCheck(fs)
        self.persister = InfoFilePersister(fs, MyLogger(backend), Suffix(randint))
        self.dir_creator = TrashDirCreator(fs)
        self.executor = JobExecutor(MyLogger(backend), TrashedFile)


    class Result(NamedTuple('Result', [
        ('ok', bool),
        ('reason', FailureReason)
    ])):
        def succeeded(self):
            return self.ok

    def trash_file_in(self,
                      candidate,  # type: Candidate
                      log_data,  # type: LogData
                      environ,  # type: Environ
                      trashee,  # type: Trashee
                      ):  # type: (...) -> Result
        secure = self.security_check.check_trash_dir_is_secure(candidate)
        if isinstance(secure, Left):
            return make_error(secure)

        can_be_used = self.trashing_checker.file_could_be_trashed_in(
            trashee, candidate, environ)
        if isinstance(can_be_used, Left):
            return make_error(can_be_used)

        dirs_creation = self.dir_creator.make_candidate_dirs(candidate)
        if isinstance(dirs_creation, Left):
            return make_error(dirs_creation)

        trashinfo_data = self.info_dir.make_trashinfo_data(
            trashee.path, candidate)
        if isinstance(trashinfo_data, Left):
            return make_error(trashinfo_data)

        persisting_job = self.persister.try_persist(trashinfo_data.value())
        trashed_file = self.executor.execute(persisting_job, log_data)
        trashed = self.trash_dir.try_trash(trashee.path, trashed_file)
        if isinstance(trashed, Left):
            return make_error(trashed)

        return make_ok()


S = TypeVar('S')


def make_error(reason):  # type: (Left[S, FailureReason]) -> Janitor.Result
    return Janitor.Result(False, reason.error())


def make_ok():  # type: () -> Janitor.Result
    return Janitor.Result(True, NoLog())



from __future__ import absolute_import

import os


def path_of_backup_copy(trashinfo_path):
    trash_dir = os.path.dirname(os.path.dirname(trashinfo_path))
    basename = os.path.basename(trashinfo_path)[:-len('.trashinfo')]
    return os.path.join(trash_dir, 'files', basename)



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from __future__ import absolute_import

import os

from trashcli.fstab.volume_of import VolumeOf


def home_trash_dir_path_from_env(environ):
    if 'XDG_DATA_HOME' in environ:
        return ['%(XDG_DATA_HOME)s/Trash' % environ]
    elif 'HOME' in environ:
        return ['%(HOME)s/.local/share/Trash' % environ]
    return []


def home_trash_dir_path_from_home(home_dir):
    return '%s/.local/share/Trash' % home_dir


def home_trash_dir(environ,
                   volume_of,  # type: VolumeOf
                   ):
    paths = home_trash_dir_path_from_env(environ)
    for path in paths:
        yield path, volume_of.volume_of(path)


def volume_trash_dir1(volume, uid):
    path = os.path.join(volume, '.Trash/%s' % uid)
    yield path, volume


def volume_trash_dir2(volume, uid):
    path = os.path.join(volume, ".Trash-%s" % uid)
    yield path, volume



import os

from six.moves import map as imap

from trashcli.put.fs.fs import Fs


class SizeCounter:
    def __init__(self,
                 fs,  # type: Fs
                 ):
        self.fs = fs

    def get_size_recursive(self, path):
        if self.fs.isfile(path):
            return self.fs.getsize(path)

        files = self.list_all_files(path)
        files_sizes = imap(self.fs.getsize, files)
        return sum(files_sizes, 0)

    def list_all_files(self,
                       path,  # type: str
                       ):
        for path, dirs, files in self.fs.walk_no_follow(path):
            for f in files:
                yield os.path.join(path, f)



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from typing import List

from trashcli.fstab.volume_of import VolumeOf
from trashcli.lib.environ import Environ
from trashcli.lib.trash_dirs import (
    volume_trash_dir1, volume_trash_dir2, home_trash_dir)
from trashcli.put.core.candidate import Candidate
from trashcli.put.core.check_type import NoCheck, TopTrashDirCheck
from trashcli.put.core.path_maker_type import PathMakerType
from trashcli.put.gate import Gate


class TrashDirectoriesFinder:
    def __init__(self,
                 fs,  # type: VolumeOf
                 ):
        self.fs = fs

    def possible_trash_directories_for(self,
                                       volume,  # type: str
                                       specific_trash_dir,  # type: str
                                       environ,  # type: Environ
                                       uid,  # type: int
                                       home_fallback,  # type: bool
                                       ):  # type: (...) -> List[Candidate]
        trash_dirs = []

        def add_home_trash(path, volume, gate):
            trash_dirs.append(
                Candidate(trash_dir_path=path,
                          volume=volume,
                          path_maker_type=PathMakerType.AbsolutePaths,
                          check_type=NoCheck,
                          gate=gate))

        def add_top_trash_dir(path, volume):
            trash_dirs.append(
                Candidate(trash_dir_path=path,
                          volume=volume,
                          path_maker_type=PathMakerType.RelativePaths,
                          check_type=TopTrashDirCheck,
                          gate=Gate.SameVolume))

        def add_alt_top_trash_dir(path, volume):
            trash_dirs.append(
                Candidate(trash_dir_path=path,
                          volume=volume,
                          path_maker_type=PathMakerType.RelativePaths,
                          check_type=NoCheck,
                          gate=Gate.SameVolume))

        if specific_trash_dir:
            path = specific_trash_dir
            volume = self.fs.volume_of(path)
            trash_dirs.append(
                Candidate(trash_dir_path=path,
                          volume=volume,
                          path_maker_type=PathMakerType.RelativePaths,
                          check_type=NoCheck,
                          gate=Gate.SameVolume))
        else:
            for path, dir_volume in home_trash_dir(environ, self.fs):
                add_home_trash(path, dir_volume, Gate.SameVolume)
            for path, dir_volume in volume_trash_dir1(volume, uid):
                add_top_trash_dir(path, dir_volume)
            for path, dir_volume in volume_trash_dir2(volume, uid):
                add_alt_top_trash_dir(path, dir_volume)
            if home_fallback:
                for path, dir_volume in home_trash_dir(environ, self.fs):
                    add_home_trash(path, dir_volume, Gate.HomeFallback)
        return trash_dirs



from enum import Enum
from typing import NamedTuple, Optional

from trashcli.lib.enum_repr import repr_for_enum


class Sort(Enum):
    ByDate = "ByDate"
    ByPath = "ByPath"
    DoNot = "DoNot"

    def __repr__(self):
        return repr_for_enum(self)


class RunRestoreArgs(
    NamedTuple('RunRestoreArgs', [
        ('path', str),
        ('sort', Sort),
        ('trash_dir', Optional[str]),
        ('overwrite', bool),
    ])):
    pass



from __future__ import absolute_import

import os

from trashcli.parse_trashinfo.parse_path import parse_path


def parse_original_location(contents, volume_path):
    path = parse_path(contents)
    return os.path.join(volume_path, path)



def protocol():
    try:
        from typing import Protocol
        return Protocol
    except ImportError as e:
        from typing_extensions import Protocol
        return Protocol


Protocol = protocol()

del protocol



from abc import abstractmethod, ABCMeta

import six
from six.moves import input as _my_input


@six.add_metaclass(ABCMeta)
class Input:
    @abstractmethod
    def read_input(self, prompt):  # type: (str) -> str
        raise NotImplementedError


class RealInput(Input):
    def read_input(self, prompt):  # type: (str) -> str
        return _my_input(prompt)


class HardCodedInput(Input):
    def __init__(self, reply=None):
        self.reply, self.exception = self._reply(reply)

    def set_reply(self, reply):
        self.reply, self.exception = self._reply(reply)

    def _reply(self, reply):
        if reply is None:
            return None, ValueError("No reply set")
        else:
            return reply, None
    def raise_exception(self, exception):
        self.exception = exception

    def read_input(self, prompt):  # type: (str) -> str
        self.used_prompt = prompt
        if self.exception:
            raise self.exception
        return self.reply

    def last_prompt(self):
        return self.used_prompt



from trashcli.fstab.volume_of import VolumeOf
from trashcli.put.fs.parent_realpath import ParentRealpathFs


class VolumeOfParent:
    def __init__(self,
                 fs,  # type: VolumeOf
                 ):
        self.fs = fs

    def volume_of_parent(self, path):
        parent_realpath = ParentRealpathFs(self.fs).parent_realpath(path)
        return self.fs.volume_of(parent_realpath)



def description(program_name, printer):
    printer.usage('Usage: %s [days]' % program_name)
    printer.summary('Purge trashed files.')
    printer.options(
        "  --version   show program's version number and exit",
        "  -h, --help  show this help message and exit")
    printer.bug_reporting()



from abc import abstractmethod
from typing import Callable, Any, Iterable

from trashcli.compat import Protocol
from trashcli.restore.args import Sort
from trashcli.restore.trashed_file import TrashedFile


def sort_files(sort,  # type: Sort
               trashed_files,  # type: Iterable[TrashedFile]
               ):  # type: (...) -> Iterable[TrashedFile]
    return sorter_for(sort).sort_files(trashed_files)


class Sorter(Protocol):
    @abstractmethod
    def sort_files(self, trashed_files,  # type: Iterable[TrashedFile]
                   ):  # type: (...) -> Iterable[TrashedFile]
        raise NotImplementedError()


class NoSorter(Sorter):
    def sort_files(self, trashed_files,  # type: Iterable[TrashedFile]
                   ):  # type: (...) -> Iterable[TrashedFile]
        return trashed_files


class SortFunction(Sorter):
    def __init__(self,
                 sort_func):  # type: (Callable[[TrashedFile], Any]) -> None
        self.sort_func = sort_func

    def sort_files(self, trashed_files,  # type: Iterable[TrashedFile]
                   ):  # type: (...) -> Iterable[TrashedFile]
        return sorted(trashed_files, key=self.sort_func)


def sorter_for(sort,  # type: Sort
               ):  # type (...) -> Sorter

    path_ranking = lambda x: x.original_location + str(x.deletion_date)
    date_rankking = lambda x: x.deletion_date
    return {
        Sort.ByPath: SortFunction(path_ranking),
        Sort.ByDate: SortFunction(date_rankking),
        Sort.DoNot: NoSorter,
    }[sort]



from trashcli.empty.clock import Clock
from trashcli.empty.older_than import older_than
from trashcli.fs import ContentsOf
from trashcli.parse_trashinfo.parse_deletion_date import parse_deletion_date


class DeleteAccordingDate:
    def __init__(self,
                 reader,  # type: ContentsOf
                 clock,  # type: Clock
                 ):
        self.reader = reader
        self.clock = clock

    def ok_to_delete(self, trashinfo_path, environ,
                     parsed_days):  # type: (str, dict, int) -> bool
        if parsed_days is None:
            return True
        else:
            contents = self.reader.contents_of(trashinfo_path)
            now_value = self.clock.get_now_value(environ)
            deletion_date = parse_deletion_date(contents)
            if deletion_date is not None:
                if older_than(parsed_days, now_value, deletion_date):
                    return True
            return False



from trashcli.fs import RealExists, RealEntriesIfDirExists
from trashcli.lib.dir_reader import DirReader


class FileSystemDirReader(DirReader,
                          RealEntriesIfDirExists,
                          RealExists,
                          ):
    pass



from __future__ import print_function
from __future__ import unicode_literals

from typing import List

from trashcli.lib.my_input import Input
from trashcli.restore.file_system import ReadCwd
from trashcli.restore.output import Output
from trashcli.restore.output_recorder import OutputRecorder
from trashcli.restore.restore_asking_the_user import RestoreAskingTheUser
from trashcli.restore.restorer import Restorer
from trashcli.restore.run_restore_action import Handler
from trashcli.restore.trashed_file import TrashedFile


class HandlerImpl(Handler):
    def __init__(self,
                 input,  # type: Input
                 cwd,  # type: ReadCwd
                 restorer,  # type: Restorer
                 output,  # type: Output
                 ):
        self.input = input
        self.cwd = cwd
        self.restorer = restorer
        self.output = output

    def handle_trashed_files(self,
                             trashed_files,  # type: List[TrashedFile]
                             overwrite,  # type: bool
                             ):
        if not trashed_files:
            self.report_no_files_found(self.cwd.getcwd_as_realpath())
        else:
            for i, trashed_file in enumerate(trashed_files):
                self.output.println("%4d %s %s" % (i,
                                                   trashed_file.deletion_date,
                                                   trashed_file.original_location))
            self.restore_asking_the_user(trashed_files, overwrite)

    def restore_asking_the_user(self, trashed_files, overwrite=False):
        my_output = OutputRecorder()
        restore_asking_the_user = RestoreAskingTheUser(self.input,
                                                       self.restorer,
                                                       my_output)
        restore_asking_the_user.restore_asking_the_user(trashed_files,
                                                        overwrite)
        my_output.apply_to(self.output)

    def report_no_files_found(self, directory):  # type: (str) -> None
        self.output.println(
            "No files trashed from current dir ('%s')" % directory)



from typing import Type


def get_permission_error_class():  # type: () -> Type[Exception]
    try:
        return PermissionError
    except NameError:
        return OSError


MyPermissionError = get_permission_error_class()



from abc import abstractmethod
from typing import NamedTuple

from trashcli.compat import Protocol
from trashcli.lib.environ import Environ
from trashcli.put.core.candidate import Candidate


class LogContext(NamedTuple('LogContext', [
    ('trashee_path', str),
    ('candidate', Candidate),
    ('environ', Environ),
])):
    def shrunk_candidate_path(self):
        return self.candidate.shrink_user(self.environ)

    def trash_dir_norm_path(self):
        return self.candidate.norm_path()

    def files_dir(self):
        return self.candidate.files_dir()


class FailureReason(Protocol):
    @abstractmethod
    def log_entries(self, context):  # type: (LogContext) -> str
        raise NotImplementedError



import argparse

from typing import List

from trashcli.empty.empty_action import EmptyActionArgs
from trashcli.empty.print_time_action import PrintTimeArgs
from trashcli.lib.environ import Environ
from trashcli.lib.print_version import PrintVersionArgs
from trashcli.shell_completion import TRASH_DIRS, add_argument_to


class Parser:
    def parse(self,
              default_is_interactive,  # type: bool
              environ,  # type: Environ
              args,  # type: List[str]
              uid,  # type: int
              argv0,  # type: str
              ):
        parser = self.make_parser(default_is_interactive)
        namespace = parser.parse_args(args)

        if namespace.version:
            return PrintVersionArgs(
                argv0=argv0,
            )
        elif namespace.print_time:
            return PrintTimeArgs(environ=environ)
        else:
            return EmptyActionArgs(
                user_specified_trash_dirs=namespace.user_specified_trash_dirs,
                all_users=namespace.all_users,
                interactive=namespace.interactive,
                days=namespace.days,
                dry_run=namespace.dry_run,
                verbose=namespace.verbose,
                environ=environ,
                uid=uid,
            )

    @staticmethod
    def make_parser(default_is_interactive):
        parser = argparse.ArgumentParser(
            description='Purge trashed files.',
            epilog='Report bugs to https://github.com/andreafrancia/trash-cli/issues')
        add_argument_to(parser)
        parser.add_argument('--version', action='store_true', default=False,
                            help="show program's version number and exit")
        parser.add_argument("-v",
                            "--verbose",
                            default=0,
                            action="count",
                            dest="verbose",
                            help="list files that will be deleted",
                            )
        parser.add_argument('--trash-dir', action='append', default=[],
                            metavar='TRASH_DIR',
                            dest='user_specified_trash_dirs',
                            help='specify the trash directory to use'
                            ).complete = TRASH_DIRS
        parser.add_argument('--print-time', action='store_true',
                            dest='print_time',
                            help=argparse.SUPPRESS)
        parser.add_argument('--all-users', action='store_true',
                            dest='all_users',
                            help='empty all trashcan of all the users')
        parser.add_argument('-i',
                            '--interactive',
                            action='store_true',
                            dest='interactive',
                            help='ask before emptying trash directories',
                            default=default_is_interactive)
        parser.add_argument('-f',
                            action='store_false',
                            help='don\'t ask before emptying trash directories',
                            dest='interactive')
        parser.add_argument('--dry-run',
                            action='store_true',
                            help='show which files would have been removed',
                            dest='dry_run')
        parser.add_argument('days', action='store', default=None, type=int,
                            nargs='?')

        return parser



import abc

from six import add_metaclass

from trashcli.fs import IsSymLink, ContentsOf, EntriesIfDirExists, PathExists, \
    IsStickyDir, HasStickyBit


@add_metaclass(abc.ABCMeta)
class FileSystemReaderForListCmd(
    IsStickyDir,
    HasStickyBit,
    IsSymLink,
    ContentsOf,
    EntriesIfDirExists,
    PathExists,
):
    pass



import os
from typing import Iterable
from trashcli.compat import Protocol

from trashcli.fs import PathExists, IsStickyDir, IsSymLink
from trashcli.fstab.volume_listing import VolumesListing
from trashcli.lib.dir_checker import DirChecker
from trashcli.lib.user_info import UserInfoProvider


class MyEnum(str):
    def __repr__(self):
        return str(self)


trash_dir_found = MyEnum('trash_dir_found')
trash_dir_skipped_because_parent_not_sticky = \
    MyEnum('trash_dir_skipped_because_parent_not_sticky')
trash_dir_skipped_because_parent_is_symlink = \
    MyEnum('trash_dir_skipped_because_parent_is_symlink')

top_trash_dir_does_not_exist = MyEnum('top_trash_dir_does_not_exist')
top_trash_dir_invalid_because_not_sticky = \
    MyEnum('top_trash_dir_invalid_because_not_sticky')
top_trash_dir_invalid_because_parent_is_symlink = \
    MyEnum('top_trash_dir_invalid_because_parent_is_symlink')
top_trash_dir_valid = MyEnum('top_trash_dir_valid')


class TrashDir(tuple):
    @property
    def path(self):
        return self[0]

    @property
    def volume(self):
        return self[1]

    def __new__(cls, path, volume):
        return tuple.__new__(TrashDir, (path, volume))

    def __repr__(self):
        return 'TrashDir(%r, %r)' % (self.path, self.volume)


class TopTrashDirRules:
    class Reader(PathExists, IsStickyDir, IsSymLink, Protocol):
        pass

    def __init__(self, reader):  # type: (Reader) -> None
        self.reader = reader

    def valid_to_be_read(self, path):
        parent_trashdir = os.path.dirname(path)
        if not self.reader.exists(path):
            return top_trash_dir_does_not_exist
        if not self.reader.is_sticky_dir(parent_trashdir):
            return top_trash_dir_invalid_because_not_sticky
        if self.reader.is_symlink(parent_trashdir):
            return top_trash_dir_invalid_because_parent_is_symlink
        else:
            return top_trash_dir_valid


class TrashDirsScanner:
    def __init__(self,
                 user_info_provider,  # type: UserInfoProvider
                 volumes_listing,  # type: VolumesListing
                 top_trash_dir_rules,  # type: TopTrashDirRules
                 dir_checker,  # type: DirChecker
                 ):
        self.user_info_provider = user_info_provider
        self.volumes_listing = volumes_listing  # type: VolumesListing
        self.top_trash_dir_rules = top_trash_dir_rules
        self.dir_checker = dir_checker

    def scan_trash_dirs(self, environ, uid):
        for user_info in self.user_info_provider.get_user_info(environ, uid):
            for path in user_info.home_trash_dir_paths:
                yield trash_dir_found, TrashDir(path, '/')
            for volume in self.volumes_listing.list_volumes(environ):
                top_trash_dir_path = os.path.join(volume, '.Trash',
                                                  str(user_info.uid))
                result = self.top_trash_dir_rules.valid_to_be_read(
                    top_trash_dir_path)
                if result == top_trash_dir_valid:
                    yield trash_dir_found, TrashDir(top_trash_dir_path, volume)
                elif result == top_trash_dir_invalid_because_not_sticky:
                    yield trash_dir_skipped_because_parent_not_sticky, (
                        top_trash_dir_path,)
                elif result == top_trash_dir_invalid_because_parent_is_symlink:
                    yield trash_dir_skipped_because_parent_is_symlink, (
                        top_trash_dir_path,)
                alt_top_trash_dir = os.path.join(volume,
                                                 '.Trash-%s' % user_info.uid)
                if self.dir_checker.is_dir(alt_top_trash_dir):
                    yield trash_dir_found, TrashDir(alt_top_trash_dir, volume)


def only_found(events, # type: Iterable[TrashDir]
               ):  # type: (...) -> Iterable[TrashDir]
    for event, args in events:
        if event == trash_dir_found:
            yield args



from typing import IO
from typing import List

from trashcli.compat import Protocol
from trashcli.put.core.logs import Level
from trashcli.put.core.logs import LogData
from trashcli.put.core.logs import LogEntry


class LoggerBackend(Protocol):
    def write_message(self,
                      log_entry,  # type: LogEntry
                      log_data,  # type: LogData
                      ):
        raise NotImplementedError()


class StreamBackend(LoggerBackend):
    def __init__(self,
                 stderr,  # type: IO[str]
                 ):  # type: (...) -> None
        self.stderr = stderr

    def write_message(self,
                      log_entry,  # type: LogEntry
                      log_data,  # type: LogData
                      ):
        if is_right_for_level(log_data.verbose, log_entry.level):
            for message in log_entry.resolve_messages():
                self.stderr.write("%s: %s\n" % (log_data.program_name, message))


def is_right_for_level(verbose,  # type: int
                       level,  # type: Level
                       ):
    min_level = {
        Level.WARNING: 0,
        Level.INFO: 1,
        Level.DEBUG: 2,
    }
    return verbose >= min_level[level]


class MyLogger:
    def __init__(self,
                 backend,  # type: LoggerBackend
                 ):  # type: (...) -> None
        self.backend = backend

    def log_put(self,
                entry,  # type: LogEntry
                log_data,  # type: LogData
                ):
        self.backend.write_message(entry, log_data)

    def log_multiple(self,
                     entries,  # type: List[LogEntry]
                     log_data,  # type: LogData
                     ):  # type: (...) -> None
        for entry in entries:
            self.log_put(entry, log_data)



import os

from trashcli.fs import file_size
from trashcli.lib.path_of_backup_copy import path_of_backup_copy
from trashcli.parse_trashinfo.maybe_parse_deletion_date import \
    maybe_parse_deletion_date


class DeletionDateExtractor:
    def extract_attribute(self, _trashinfo_path, contents):
        return maybe_parse_deletion_date(contents)


class SizeExtractor:
    def extract_attribute(self, trashinfo_path, _contents):
        backup_copy = path_of_backup_copy(trashinfo_path)
        try:
            return str(file_size(backup_copy))
        except FileNotFoundError:
            if os.path.islink(backup_copy):
                return 0
            else:
                raise



def octal(n):  # type: (int) -> str
    return "%s%s" % ("0o", _remove_octal_prefix(oct(n)))


def _remove_octal_prefix(o):
    if o.startswith('0o'):
        return o[2:]
    elif o.startswith('0'):
        return o[1:]
    else:
        ValueError('Invalid octal format: %s' % o)



from typing import Any
from typing import Type
from typing import TypeVar
from typing import cast

T = TypeVar('T')


def check_cast(t, value):  # type: (Type[T], Any) -> T
    if isinstance(value, t):
        return cast(t, value)  # type: ignore
    else:
        raise TypeError("expected %s, got %s" % (t, type(value)))



import os

from trashcli.put.fs.fs import RealPathFs


class TrashDirVolumeReader:
    def __init__(self,
                 fs,  # type: RealPathFs
                 ):
        self.fs = fs

    def volume_of_trash_dir(self, trash_dir_path):
        norm_trash_dir_path = os.path.normpath(trash_dir_path)
        return self.fs.volume_of(
            self.fs.realpath(norm_trash_dir_path))



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from abc import abstractmethod, ABCMeta

import six
from typing import Optional

from trashcli.fstab.volume_of import VolumeOf
from trashcli.fstab.volumes import Volumes
from trashcli.lib.environ import Environ
from trashcli.lib.trash_dirs import (
    volume_trash_dir1, volume_trash_dir2, home_trash_dir)


@six.add_metaclass(ABCMeta)
class TrashDirectories:
    @abstractmethod
    def list_trash_dirs(self, trash_dir_from_cli):
        raise NotImplementedError()


class TrashDirectoriesImpl(TrashDirectories):
    def __init__(self,
                 volumes,  # type: Volumes
                 uid,  # type: int
                 environ,
                 ):
        trash_directories1 = TrashDirectories1(volumes, uid, environ)
        self.trash_directories2 = TrashDirectories2(volumes,
                                                    trash_directories1)

    def list_trash_dirs(self,
                        trash_dir_from_cli,  # type: Optional[str]
                        ):
        return self.trash_directories2.trash_directories_or_user(
            trash_dir_from_cli)


class TrashDirectories2:
    def __init__(self,
                 volume_of,  # type: VolumeOf
                 trash_directories,  # type: TrashDirectories1
                 ):
        self.volume_of = volume_of
        self.trash_directories = trash_directories

    def trash_directories_or_user(self,
                                  trash_dir_from_cli,  # type: Optional[str]
                                  ):
        if trash_dir_from_cli:
            return [(trash_dir_from_cli,
                     self.volume_of.volume_of(trash_dir_from_cli))]
        return self.trash_directories.all_trash_directories()



class TrashDirectories1:
    def __init__(self,
                 volumes,  # type: Volumes
                 uid,  # type: int
                 environ,  # type: Environ
                 ):
        self.volumes = volumes
        self.uid = uid
        self.environ = environ

    def all_trash_directories(self):
        volumes_to_check = self.volumes.list_mount_points()
        for path1, volume1 in home_trash_dir(self.environ, self.volumes):
            yield path1, volume1
        for volume in volumes_to_check:
            for path1, volume1 in volume_trash_dir1(volume, self.uid):
                yield path1, volume1
            for path1, volume1 in volume_trash_dir2(volume, self.uid):
                yield path1, volume1



# Copyright (C) 2022 Andrea Francia Bereguardo(PV) Italy
from typing import Iterable

from trashcli.empty.console import Console
from trashcli.empty.delete_according_date import DeleteAccordingDate
from trashcli.empty.existing_file_remover import ExistingFileRemover
from trashcli.lib.path_of_backup_copy import path_of_backup_copy
from trashcli.lib.trash_dir_reader import TrashDirReader
from trashcli.trash_dirs_scanner import TrashDir, only_found


class Emptier:
    def __init__(self, delete_mode, trash_dir_reader, file_remover, console
                 ):  # type: (DeleteAccordingDate, TrashDirReader, ExistingFileRemover, Console) -> None
        self.console = console
        self.file_remover = file_remover
        self.delete_mode = delete_mode
        self.trash_dir_reader = trash_dir_reader

    def do_empty(self,
                 trash_dirs,  # type: Iterable[TrashDir]
                 environ,  # type: dict
                 parsed_days,  # type: int
                 dry_run,  # type: bool
                 verbose,  # type: int
                 ):  # type: (...) -> None
        for path in self.files_to_delete(trash_dirs, environ, parsed_days):
            if dry_run:
                self.console.print_dry_run(path)
            else:
                if verbose:
                    self.console.print_removing(path)
                try:
                    self.file_remover.remove_file_if_exists(path)
                except OSError:
                    self.console.print_cannot_remove_error(path)

    def files_to_delete(self,
                        trash_dirs,  # type: Iterable[TrashDir]
                        environ,  # type: dict
                        parsed_days,  # type: int
                        ):  # type: (...) -> Iterable[str]
        for trash_dir in only_found(trash_dirs):  # type: TrashDir
            for trash_info_path in self.trash_dir_reader.list_trashinfo(
                    trash_dir.path):
                if self.delete_mode.ok_to_delete(trash_info_path, environ,
                                                 parsed_days):
                    yield (path_of_backup_copy(trash_info_path))
                    yield trash_info_path
            for orphan in self.trash_dir_reader.list_orphans(
                    trash_dir.path):
                yield orphan



# Copyright (C) 2011-2022 Andrea Francia Bereguardo(PV) Italy
import os
import sys

import trashcli.trash
from trashcli.empty.main import ContentReader
from trashcli.file_system_reader import FileSystemReader
from trashcli.fs import RealContentsOf
from trashcli.fstab.volume_listing import RealVolumesListing
from trashcli.fstab.real_volume_of import RealVolumeOf
from trashcli.fstab.volume_of import VolumeOf
from trashcli.lib.dir_reader import DirReader, RealDirReader
from trashcli.lib.print_version import PrintVersionArgs, \
    PrintVersionAction
from trashcli.list.list_trash_action import ListTrashAction, ListTrashArgs
from trashcli.list.minor_actions.debug_volumes import DebugVolumes, \
    DebugVolumesArgs
from trashcli.list.minor_actions.list_trash_dirs import ListTrashDirs, \
    ListTrashDirsArgs
from trashcli.list.minor_actions.list_volumes import PrintVolumesList, \
    PrintVolumesArgs
from trashcli.list.minor_actions.print_python_executable import \
    PrintPythonExecutable, PrintPythonExecutableArgs
from trashcli.list.parser import Parser
from trashcli.list.trash_dir_selector import TrashDirsSelector
from trashcli.trash_dirs_scanner import TopTrashDirRules


def main():
    ListCmd(
        out=sys.stdout,
        err=sys.stderr,
        environ=os.environ,
        volumes_listing=RealVolumesListing(),
        uid=os.getuid(),
        volumes=RealVolumeOf(),
        dir_reader=RealDirReader(),
        file_reader=FileSystemReader(),
        content_reader=RealContentsOf(),
        version=trashcli.trash.version
    ).run(sys.argv)


class ListCmd:
    def __init__(self,
                 out,
                 err,
                 environ,
                 volumes_listing,
                 uid,
                 volumes,  # type: VolumeOf
                 file_reader, # type: TopTrashDirRules.Reader
                 dir_reader,  # type: DirReader
                 content_reader, # type: ContentReader
                 version,
                 ):
        self.out = out
        self.err = err
        self.version = version
        self.dir_reader = dir_reader
        self.content_reader = content_reader
        self.environ = environ
        self.uid = uid
        self.volumes_listing = volumes_listing
        self.selector = TrashDirsSelector.make(volumes_listing,
                                               file_reader,
                                               volumes)
        self.actions = {PrintVersionArgs: PrintVersionAction(self.out,
                                                             self.version),
                        PrintVolumesArgs: PrintVolumesList(self.environ,
                                                           self.volumes_listing,
                                                           self.out),
                        DebugVolumesArgs: DebugVolumes(),
                        ListTrashDirsArgs: ListTrashDirs(self.environ,
                                                         self.uid,
                                                         self.selector),
                        ListTrashArgs: ListTrashAction(self.environ,
                                                       self.uid,
                                                       self.selector,
                                                       self.out,
                                                       self.err,
                                                       self.dir_reader,
                                                       self.content_reader),
                        PrintPythonExecutableArgs: PrintPythonExecutable()}

    def run(self, argv):
        parser = Parser(os.path.basename(argv[0]))
        args = parser.parse_list_args(argv[1:], argv[0])

        action = self.actions[type(args)]

        action.run_action(args)



from typing import NamedTuple, List

from trashcli.empty.clock import Clock
from trashcli.empty.console import Console
from trashcli.empty.delete_according_date import (
    DeleteAccordingDate,
)
from trashcli.empty.emptier import Emptier
from trashcli.empty.existing_file_remover import ExistingFileRemover
from trashcli.empty.guard import Guard
from trashcli.empty.parse_reply import parse_reply
from trashcli.empty.prepare_output_message import prepare_output_message
from trashcli.empty.user import User
from trashcli.fs import ContentsOf
from trashcli.fstab.volume_listing import VolumesListing
from trashcli.fstab.volume_of import VolumeOf
from trashcli.lib.dir_reader import DirReader
from trashcli.lib.environ import Environ
from trashcli.lib.my_input import RealInput
from trashcli.lib.trash_dir_reader import TrashDirReader
from trashcli.list.trash_dir_selector import TrashDirsSelector
from trashcli.trash_dirs_scanner import TopTrashDirRules


class EmptyActionArgs(
    NamedTuple('EmptyActionArgs', [
        ('user_specified_trash_dirs', List[str]),
        ('all_users', bool),
        ('interactive', bool),
        ('days', int),
        ('dry_run', bool),
        ('verbose', int),
        ('environ', Environ),
        ('uid', int),
    ])):
    pass


class EmptyAction:
    def __init__(self,
                 clock,  # type: Clock
                 file_remover,  # type: ExistingFileRemover
                 volumes_listing,  # type: VolumesListing
                 file_reader,  # type: TopTrashDirRules.Reader
                 volumes,  # type: VolumeOf
                 dir_reader,  # type: DirReader
                 content_reader,  # type: ContentsOf
                 console,  # type: Console
                 ):  # type: (...) -> None
        self.selector = TrashDirsSelector.make(volumes_listing,
                                               file_reader,
                                               volumes)
        trash_dir_reader = TrashDirReader(dir_reader)
        delete_mode = DeleteAccordingDate(content_reader,
                                          clock)
        user = User(prepare_output_message, RealInput(), parse_reply)
        self.emptier = Emptier(delete_mode, trash_dir_reader, file_remover,
                               console)
        self.guard = Guard(user)

    def run_action(self,
                   args,  # type: EmptyActionArgs
                   ):  # type: (...) -> None
        trash_dirs = self.selector.select(args.all_users,
                                          args.user_specified_trash_dirs,
                                          args.environ,
                                          args.uid)
        delete_pass = self.guard.ask_the_user(args.interactive,
                                              trash_dirs)
        if delete_pass.ok_to_empty:
            self.emptier.do_empty(delete_pass.trash_dirs, args.environ,
                                  args.days, args.dry_run, args.verbose)



from trashcli.put.my_logger import LoggerBackend

from trashcli.put.context import Context
from trashcli.put.context import SingleTrasher
from trashcli.put.core.trash_result import TrashResult
from trashcli.put.core.trashee import should_skipped_by_specs
from trashcli.put.file_trasher import FileTrasher
from trashcli.put.fs.fs import Fs
from trashcli.put.my_logger import MyLogger
from trashcli.put.reporting.trash_put_reporter import TrashPutReporter
from trashcli.put.user import User
from trashcli.put.user import user_replied_no


class Trasher(SingleTrasher):
    def __init__(self,
                 file_trasher,  # type: FileTrasher
                 user,  # type: User
                 fs,  # type: Fs
                 backend,  # type: LoggerBackend
                 ):
        self.file_trasher = file_trasher
        self.user = user
        self.logger = MyLogger(backend)
        self.reporter = TrashPutReporter(fs)
        self.fs = fs

    def trash_single(self,
                     path,  # type: str
                     context,  # type: Context
                     ):
        """
        Trash a file in the appropriate trash directory.
        If the file belong to the same volume of the trash home directory it
        will be trashed in the home trash directory.
        Otherwise it will be trashed in one of the relevant volume trash
        directories.

        Each volume can have two trash directories, they are
            - $volume/.Trash/$uid
            - $volume/.Trash-$uid

        Firstly the software attempt to trash the file in the first directory
        then try to trash in the second trash directory.
        """

        if should_skipped_by_specs(path):
            self.logger.log_put(
                self.reporter.unable_to_trash_dot_entries(path),
                context.log_data)
            return TrashResult.Failure

        if not self.fs.lexists(path):
            if context.mode.can_ignore_not_existent_path():
                return TrashResult.Success
            else:
                self.logger.log_put(
                    self.reporter.unable_to_trash_file_non_existent(path),
                    context.log_data)
                return TrashResult.Failure

        if context.mode.should_we_ask_to_the_user(self.fs.is_accessible(path)):
            reply = self.user.ask_user_about_deleting_file(context.program_name,
                                                           path)
            if reply == user_replied_no:
                return TrashResult.Success

        return self.file_trasher.trash_file(path, context)



from typing import TextIO

from trashcli.empty.errors import format_error_msg


class Console:
    def __init__(self, program_name, out,
                 err):  # type: (str, TextIO, TextIO) -> None
        self.program_name = program_name
        self.out = out
        self.err = err

    def print_cannot_remove_error(self, path):
        self.print_error("cannot remove %s" % path)

    def print_error(self, msg):
        self.err.write(format_error_msg(self.program_name, msg))

    def print_dry_run(self, path):
        self.out.write("would remove %s\n" % path)

    def print_removing(self, path):
        self.out.write("removing %s\n" % path)



import os
from datetime import datetime
from typing import TextIO, Callable

from trashcli.empty.clock import Clock
from trashcli.empty.console import Console
from trashcli.empty.empty_action import EmptyAction, EmptyActionArgs
from trashcli.empty.errors import Errors
from trashcli.empty.existing_file_remover import ExistingFileRemover
from trashcli.empty.is_input_interactive import is_input_interactive
from trashcli.empty.parser import Parser
from trashcli.empty.print_time_action import PrintTimeAction, PrintTimeArgs
from trashcli.fs import ContentsOf
from trashcli.fstab.volume_listing import VolumesListing
from trashcli.fstab.volume_of import VolumeOf
from trashcli.lib.dir_reader import DirReader
from trashcli.lib.exit_codes import EX_OK
from trashcli.lib.print_version import PrintVersionAction, PrintVersionArgs
from trashcli.trash_dirs_scanner import TopTrashDirRules


class EmptyCmd:
    def __init__(self,
                 argv0,  # type: str
                 out,  # type: TextIO
                 err,  # type: TextIO
                 volumes_listing,  # type: VolumesListing
                 now,  # type: Callable[[], datetime]
                 file_reader,  # type: TopTrashDirRules.Reader
                 dir_reader,  # type: DirReader
                 content_reader,  # type: ContentsOf
                 file_remover,  # type: ExistingFileRemover
                 version,  # type: str
                 volumes,  # type: VolumeOf
                 ):
        self.volumes = volumes
        self.file_remover = file_remover
        self.dir_reader = dir_reader
        self.file_reader = file_reader
        self.volumes_listing = volumes_listing
        self.argv0 = argv0
        self.out = out
        self.err = err
        self.version = version
        self.now = now
        self.content_reader = content_reader
        self.parser = Parser()
        self.program_name = os.path.basename(argv0)
        errors = Errors(self.program_name, self.err)
        clock = Clock(self.now, errors)
        console = Console(self.program_name, self.out, self.err)
        self.empty_action = EmptyAction(clock,
                                        self.file_remover,
                                        self.volumes_listing,
                                        self.file_reader,
                                        self.volumes,
                                        self.dir_reader,
                                        self.content_reader,
                                        console)
        self.print_version_action = PrintVersionAction(self.out,
                                                       self.version)
        self.print_time_action = PrintTimeAction(self.out, clock)

    def run_cmd(self, args, environ, uid):
        args = self.parser.parse(
            default_is_interactive=is_input_interactive(),
            args=args,
            argv0=self.argv0,
            environ=environ,
            uid=uid)

        if type(args) is PrintVersionArgs:
            return self.print_version_action.run_action(args)
        elif type(args) is EmptyActionArgs:
            return self.empty_action.run_action(args)
        elif type(args) is PrintTimeArgs:
            return self.print_time_action.run_action(args)

        return EX_OK



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
import logging

my_logger = logging.getLogger('trashcli.trash')
my_logger.setLevel(logging.WARNING)
my_logger.addHandler(logging.StreamHandler())



# Copyright (C) 2007-2024 Andrea Francia Trivolzio(PV) Italy
import os
from typing import List
from typing import Tuple

from trashcli.put.fs.fs import Fs

from trashcli.lib.environ import Environ
from trashcli.lib.exit_codes import EX_IOERR
from trashcli.lib.exit_codes import EX_OK
from trashcli.put.core.candidate import Candidate
from trashcli.put.core.failure_reason import FailureReason
from trashcli.put.core.failure_reason import LogContext
from trashcli.put.core.logs import Level
from trashcli.put.core.logs import LogEntry
from trashcli.put.core.logs import LogTag
from trashcli.put.core.logs import debug_str
from trashcli.put.core.logs import info_str
from trashcli.put.core.logs import log_str
from trashcli.put.core.trash_all_result import TrashAllResult
from trashcli.put.core.trashee import Trashee
from trashcli.put.describer import Describer
from trashcli.put.reporting.stats_reader import gentle_stat_read


class TrashPutReporter:
    def __init__(self,
                 fs,  # type: Fs
                 ):
        self.describer = Describer(fs)

    def _describe(self, path):
        return self.describer.describe(path)

    def unable_to_trash_dot_entries(self, file
                                    ):  # type: (...) -> LogEntry
        return log_str(Level.WARNING, LogTag.trash_failed,
                       ["cannot trash %s '%s'" % (self._describe(file), file)])

    def unable_to_trash_file_non_existent(self,
                                          path,  # type: str
                                          ):  # type: (...) -> LogEntry
        return log_str(Level.WARNING, LogTag.trash_failed,
                       ["cannot trash %s '%s'" % (self._describe(path), path)])

    def unable_to_trash_file(
            self,
            trashee,  # type: Trashee
            failures,  # type: List[Tuple[Candidate, FailureReason]]
            environ,  # type: Environ
    ):  # (...) -> LogEntry
        path = trashee.path
        volume = trashee.volume
        messages = []
        messages.append("cannot trash %s '%s' (from volume '%s')" % (
            self._describe(path), path, volume))
        for candidate, reason in failures:
            context = LogContext(path, candidate, environ)
            message = " `- failed to trash %s in %s, because %s" % (
                path,
                candidate.norm_path(),
                reason.log_entries(context))
            messages.append(message)

        return log_str(Level.WARNING, LogTag.trash_failed, messages)

    @staticmethod
    def file_has_been_trashed_in_as(trashed_file,
                                    trash_dir,  # type: Candidate
                                    environ):  # type: (...) -> LogEntry
        return info_str("'%s' trashed in %s" % (trashed_file,
                                                trash_dir.shrink_user(environ)))

    @classmethod
    def log_data_for_debugging(cls, error):
        try:
            filename = error.filename
        except AttributeError:
            pass
        else:
            if filename is not None:
                for path in [filename, os.path.dirname(filename)]:
                    info = gentle_stat_read(path)
                    yield "stats for %s: %s" % (path, info)

    @staticmethod
    def trash_dir_with_volume(candidate,  # type: Candidate
                              ):  # type: (...) -> LogEntry
        return debug_str("trying trash dir: %s from volume: %s" % (
            candidate.norm_path(), candidate.volume))

    @staticmethod
    def exit_code(result,  # type: TrashAllResult
                  ):  # type: (...) -> int
        if not result.any_failure():
            return EX_OK
        else:
            return EX_IOERR



from typing import Generic
from typing import Iterator
from typing import Type
from typing import TypeVar

from trashcli.put.core.logs import Level
from trashcli.put.core.logs import LogData
from trashcli.put.core.logs import LogEntry
from trashcli.put.core.logs import LogTag
from trashcli.put.core.logs import MessageStr
from trashcli.put.my_logger import MyLogger

R = TypeVar('R')


class JobStatus(Generic[R]):
    def __init__(self, message):  # type: (str) -> None
        self.log_entries = [LogEntry(Level.DEBUG,
                                     LogTag.unspecified,
                                     MessageStr.from_messages([message]))]

    def has_succeeded(self):
        return isinstance(self, Succeeded)

    def result(self):  # type: () -> R
        raise NotImplementedError

    def logs(self):
        return self.log_entries


class NeedsMoreAttempts(JobStatus, Generic[R]):
    def __init__(self,
                 trashinfo_path,  # type: str
                 message,  # type: str
                 ):
        super(NeedsMoreAttempts, self).__init__(message)
        self.trashinfo_path = trashinfo_path

    def result(self):  # type: () -> R
        raise ValueError("Result not available yet!")


class Succeeded(JobStatus, Generic[R]):
    def __init__(self,
                 result,  # type: R
                 message,  # type: str
                 ):
        super(Succeeded, self).__init__(message)
        self._result = result  # type: R

    def result(self):  # type: () -> R
        return self._result

    def __repr__(self):
        return "Succeeded(%s)" % repr(self.result)


class JobExecutor(Generic[R]):
    def __init__(self,
                 logger,  # type: MyLogger
                 _result_type,  # type: Type[R]
                 ):
        self.logger = logger

    def execute(self,
                job,  # type: Iterator[JobStatus[R]]
                log_data,  # type: LogData
                ):  # type: (...) -> R
        for status in job:
            self.logger.log_multiple(status.logs(), log_data)
            if status.has_succeeded():
                return status.result()
        raise ValueError("Should not happen!")



# Copyright (C) 2011-2022 Andrea Francia Bereguardo(PV) Italy
import os
import sys
from datetime import datetime

from trashcli.compat import Protocol

from trashcli import trash
from trashcli.empty.empty_cmd import EmptyCmd
from trashcli.fs import RealContentsOf, ContentsOf
from .existing_file_remover import ExistingFileRemover
from .file_system_dir_reader import FileSystemDirReader
from .top_trash_dir_rules_file_system_reader import \
    RealTopTrashDirRulesReader
from ..fstab.volume_listing import RealVolumesListing
from ..fstab.real_volume_of import RealVolumeOf


class ContentReader(ContentsOf, Protocol):
    pass


def main():
    empty_cmd = EmptyCmd(argv0=sys.argv[0],
                         out=sys.stdout,
                         err=sys.stderr,
                         volumes_listing=RealVolumesListing(),
                         now=datetime.now,
                         file_reader=RealTopTrashDirRulesReader(),
                         file_remover=ExistingFileRemover(),
                         content_reader=FileSystemContentReader(),
                         dir_reader=FileSystemDirReader(),
                         version=trash.version,
                         volumes=RealVolumeOf())
    return empty_cmd.run_cmd(sys.argv[1:], os.environ, os.getuid())


class FileSystemContentReader(ContentReader, RealContentsOf):
    pass



from typing import NamedTuple, Union

Quit = NamedTuple('Quit', [])
Die = NamedTuple('Die', [('msg', Union[str, Exception])])
Println = NamedTuple('Println', [('msg', str)])
Exiting = NamedTuple('Exiting', [('msg', str)])
OutputEvent = Union[Quit, Die, Println, Exiting]



from trashcli.fs import RealIsStickyDir, RealHasStickyBit, \
    RealIsSymLink, RealContentsOf, RealEntriesIfDirExists, RealExists
from trashcli.list.fs import FileSystemReaderForListCmd


class FileSystemReader(FileSystemReaderForListCmd,
                       RealIsStickyDir,
                       RealHasStickyBit,
                       RealIsSymLink,
                       RealContentsOf,
                       RealEntriesIfDirExists,
                       RealExists
                       ):
    pass



from typing import List

from trashcli.lib.environ import Environ
from trashcli.put.context import Context
from trashcli.put.my_logger import LogData
from trashcli.put.parser import ExitWithCode
from trashcli.put.parser import Parser
from trashcli.put.parser import Trash
from trashcli.put.reporting.trash_put_reporter import TrashPutReporter
from trashcli.put.trasher import Trasher


class TrashPutCmd:
    def __init__(self,
                 reporter,  # type: TrashPutReporter
                 trasher,  # type: Trasher
                 ):
        self.reporter = reporter
        self.trasher = trasher

    def run_put(self,
                argv,  # type: List[str]
                environ,  # type: Environ
                uid,  # type: int
                ):  # type: (...) -> int
        parser = Parser()
        parsed = parser.parse_args(argv)
        if isinstance(parsed, ExitWithCode):
            return parsed.exit_code
        elif isinstance(parsed, Trash):
            program_name = parsed.program_name
            log_data = LogData(program_name, parsed.verbose)
            context = Context(paths=parsed.files,
                              user_trash_dir=parsed.trash_dir,
                              mode=parsed.mode,
                              forced_volume=parsed.forced_volume,
                              home_fallback=parsed.home_fallback,
                              program_name=program_name,
                              log_data=log_data,
                              environ=environ,
                              uid=uid)
            result = context.trash_each(self.trasher)

            return self.reporter.exit_code(result)



import os
from typing import NamedTuple


class Trashee(NamedTuple('FileToBeTrashed', [
    ('path', str),
    ('volume', str)
])):
    pass


def should_skipped_by_specs(path):
    basename = os.path.basename(path)
    return (basename == ".") or (basename == "..")



import os
from typing import Union, List, cast

from trashcli.lib.print_version import PrintVersionArgs
from trashcli.restore.args import RunRestoreArgs, Sort
from trashcli.shell_completion import add_argument_to, TRASH_FILES, TRASH_DIRS, \
    complete_with

Command = Union[PrintVersionArgs, RunRestoreArgs]


class RestoreArgParser:
    def parse_restore_args(self,
                           sys_argv,  # type: List[str]
                           curdir,  # type: str
                           ):  # type: (...) -> Command
        import argparse

        parser = argparse.ArgumentParser(
            description='Restores from trash chosen file',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        add_argument_to(parser)
        complete_with(
            TRASH_FILES,
            parser.add_argument('path',
                                default="", nargs='?',
                                help='Restore files from given path instead of current '
                                     'directory'
                                )
        )
        parser.add_argument('--sort',
                            choices=['date', 'path', 'none'],
                            default='date',
                            help='Sort list of restore candidates by given field')
        complete_with(
            TRASH_DIRS,
            parser.add_argument('--trash-dir',
                                action='store',
                                dest='trash_dir',
                                help=argparse.SUPPRESS
                                ))
        parser.add_argument('--version', action='store_true', default=False)

        parser.add_argument('--overwrite',
                            action='store_true',
                            default=False,
                            help='Overwrite existing files with files coming out of the trash')

        parsed = parser.parse_args(sys_argv[1:])

        if parsed.version:
            return PrintVersionArgs(argv0=sys_argv[0])
        else:
            path = os.path.normpath(
                os.path.join(curdir + os.path.sep, parsed.path))

            return RunRestoreArgs(path=path,
                                  sort=cast(Sort, {
                                      'path': Sort.ByPath,
                                      'date': Sort.ByDate,
                                      'none': Sort.DoNot
                                  }[parsed.sort]),
                                  trash_dir=parsed.trash_dir,
                                  overwrite=parsed.overwrite)



# Copyright (C) 2007-2024 Andrea Francia Trivolzio(PV) Italy

import os
from grp import getgrgid
from pwd import getpwuid
from typing import NamedTuple
import re

from trashcli.put.core.either import Either
from trashcli.put.core.either import Left
from trashcli.put.core.either import Right


def gentle_stat_read(path):
    def stats_str(stats):  # type: (Either[Stats, Exception]) -> str
        if isinstance(result, Right):
            value = result.value()
            return "%s %s %s" % (value.octal_mode(), value.user, value.group)
        elif isinstance(result, Left):
            return str(result.error())
        else:
            raise ValueError()

    result = StatReader().read_stats(path)
    return stats_str(result)


class Stats(NamedTuple('Result', [
    ('user', str),
    ('group', str),
    ('mode', int),
])):
    def octal_mode(self):  # () -> str
        return self._remove_octal_prefix(oct(self.mode & 0o777))

    @staticmethod
    def _remove_octal_prefix(mode):  # type: (str) -> str
        remove_new_octal_format = mode.replace('0o', '')
        remove_old_octal_format = re.sub(r"^0", '', remove_new_octal_format)
        return remove_old_octal_format


class StatReader:
    @staticmethod
    def read_stats(path,  # type: str
                   ):  # type: (...) -> Either[Stats, Exception]
        try:
            stats = os.lstat(path)
            user = getpwuid(stats.st_uid).pw_name
            group = getgrgid(stats.st_gid).gr_name
            mode = stats.st_mode

            return Right(Stats(user, group, mode))
        except (IOError, OSError) as e:
            return Left(e)



from __future__ import print_function


class PrintVolumesArgs(object):
    pass


class PrintVolumesList(object):
    def __init__(self, environ, volumes_listing, out):
        self.environ = environ
        self.volumes_listing = volumes_listing
        self.out = out

    def run_action(self,
                   args,  # type: PrintVolumesArgs
                   ):
        for volume in self.volumes_listing.list_volumes(self.environ):
            print(volume, file=self.out)



from __future__ import absolute_import

import datetime


class Clock:
    def __init__(self, real_now, errors):
        self.real_now = real_now
        self.errors = errors

    def get_now_value(self, environ):
        if 'TRASH_DATE' in environ:
            try:
                return datetime.datetime.strptime(environ['TRASH_DATE'],
                                                  "%Y-%m-%dT%H:%M:%S")
            except ValueError:
                self.errors.print_error('invalid TRASH_DATE: %s' %
                                        environ['TRASH_DATE'])
                return self.real_now()
        return self.real_now()



import os
import pprint
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from argparse import SUPPRESS
from typing import Any
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Union

from trashcli.put.core.mode import Mode
from trashcli.shell_completion import TRASH_DIRS
from trashcli.shell_completion import TRASH_FILES
from trashcli.shell_completion import add_argument_to
from trashcli.trash import version

ExitWithCode = NamedTuple('ExitWithCode', [
    ('type', type),
    ('exit_code', int),
])

Trash = NamedTuple('Trash', [
    ('type', type),
    ('program_name', str),
    ('options', Any),
    ('files', List[str]),
    ('trash_dir', Optional[str]),
    ('mode', Mode),
    ('forced_volume', Optional[str]),
    ('verbose', int),
    ('home_fallback', bool),
])


class Parser:

    def parse_args(self, argv):  # type: (list) -> Union[ExitWithCode, Trash]
        program_name = os.path.basename(argv[0])
        arg_parser = make_parser(program_name)
        try:
            options = arg_parser.parse_args(argv[1:])
            if len(options.files) <= 0:
                arg_parser.error("Please specify the files to trash.")
        except SystemExit as e:
            return ExitWithCode(type=ExitWithCode, exit_code=ensure_int(e.code))

        return Trash(type=Trash,
                     program_name=program_name,
                     options=options,
                     files=options.files,
                     trash_dir=options.trashdir,
                     mode=options.mode,
                     forced_volume=options.forced_volume,
                     verbose=options.verbose,
                     home_fallback=options.home_fallback)


def ensure_int(code):
    if not isinstance(code, int):
        raise ValueError("code must be an int, got %s" % pprint.pformat(code))
    return code


def make_parser(program_name):
    parser = ArgumentParser(prog=program_name,
                            usage="%(prog)s [OPTION]... FILE...",
                            description="Put files in trash",
                            formatter_class=RawDescriptionHelpFormatter,
                            epilog="""\
all trash-cli commands:
  trash-put             trash files and directories.
  trash-empty           empty the trashcan(s).
  trash-list            list trashed files.
  trash-restore         restore a trashed file.
  trash-rm              remove individual files from the trashcan

To remove a file whose name starts with a '-', for example '-foo',
use one of these commands:

    trash -- -foo

    trash ./-foo

Report bugs to https://github.com/andreafrancia/trash-cli/issues""")
    add_argument_to(parser)
    parser.add_argument("-d", "--directory",
                        action="store_true",
                        help="ignored (for GNU rm compatibility)")
    parser.add_argument("-f", "--force",
                        action="store_const",
                        dest="mode",
                        const=Mode.mode_force,
                        default=Mode.mode_unspecified,
                        help="silently ignore nonexistent files")
    parser.add_argument("-i", "--interactive",
                        action="store_const",
                        dest="mode",
                        const=Mode.mode_interactive,
                        default=Mode.mode_unspecified,
                        help="prompt before every removal")
    parser.add_argument("-r", "-R", "--recursive",
                        action="store_true",
                        help="ignored (for GNU rm compatibility)")
    parser.add_argument("--trash-dir",
                        type=str,
                        action="store", dest='trashdir',
                        help='use TRASHDIR as trash folder'
                        ).complete = TRASH_DIRS
    parser.add_argument("-v",
                        "--verbose",
                        default=0,
                        action="count",
                        dest="verbose",
                        help="explain what is being done")
    parser.add_argument('--force-volume',
                        default=None,
                        action="store",
                        dest="forced_volume",
                        help=SUPPRESS)
    parser.add_argument('--home-fallback',
                        default=False,
                        action="store_true",
                        dest="home_fallback",
                        help=SUPPRESS)
    parser.add_argument("--version",
                        action="version",
                        version=version)
    parser.add_argument('files',
                        nargs='*'
                        ).complete = TRASH_FILES
    return parser



import datetime
import os
from typing import NamedTuple, Optional


class TrashedFile(
    NamedTuple('TrashedFile', [
        ('original_location', str),
        ('deletion_date', Optional[datetime.datetime]),
        ('info_file', str),
        ('original_file', str),
    ])):
    """
    Represent a trashed file.
    Each trashed file is persisted in two files:
     - $trash_dir/info/$id.trashinfo
     - $trash_dir/files/$id

    Properties:
     - path : the original path from where the file has been trashed
     - deletion_date : the time when the file has been trashed (instance of
                       datetime)
     - info_file : the file that contains information (instance of Path)
     - original_file : the path where the trashed file has been placed after the
                       trash operation (instance of Path)
    """

    def original_location_matches_path(self, path):
        if path == os.path.sep:
            return True
        if self.original_location.startswith(path + os.path.sep):
            return True
        return self.original_location == path



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy

from __future__ import absolute_import
from __future__ import print_function

import os

import six
from typing import NamedTuple


class PrintVersionArgs(
    NamedTuple('PrintVersionArgs', [
        ('argv0', str),
    ])):

    def program_name(self):
        return os.path.basename(self.argv0)


class PrintVersionAction(object):
    def __init__(self, out, version):
        self.out = out
        self.version = version

    def run_action(self,
                   args,  # type: PrintVersionArgs
                   ):
        print_version(self.out, args.program_name(), self.version)


def print_version(out, program_name, version):
    print("%s %s" % (program_name, six.text_type(version)), file=out)



class PrintPythonExecutableArgs:
    pass


class PrintPythonExecutable:
    def run_action(self,
                   _args,  # type: PrintPythonExecutableArgs
                   ):
        import sys
        print(sys.executable)



from abc import abstractmethod
from enum import Enum
from typing import List
from typing import NamedTuple
from trashcli.compat import Protocol


class Level(Enum):
    INFO = "INFO"
    DEBUG = "DEBUG"
    WARNING = "WARNING"


class LogTag(Enum):
    "tags used only during testing"
    unspecified = "unspecified"
    trash_failed = "trash_failed"


class LogData:
    def __init__(self, program_name, verbose):
        self.program_name = program_name
        self.verbose = verbose


class Message(Protocol):
    @abstractmethod
    def resolve(self):
        raise NotImplementedError


class LogEntry(NamedTuple('LogEntry', [
    ('level', Level),
    ('tag', LogTag),
    ('messages', List[Message]),
])):
    def resolve_messages(self):
        for m in self.messages:
            yield m.resolve()


class MessageStr(NamedTuple('Message', [
    ('message', str),
]), Message):
    def resolve(self):
        return self.message

    @staticmethod
    def from_messages(messages  # type: List[str]
                      ):
        return [MessageStr(msg) for msg in messages]


def log_str(level,  # type: Level
            tag,  # type: LogTag
            messages,  # type: List[str]
            ):
    return LogEntry(level, tag, MessageStr.from_messages(messages))


def warning_str(message):  # type: (str) -> LogEntry
    return log_str(Level.WARNING, LogTag.unspecified, [message])


def info_str(message):  # type: (str) -> LogEntry
    return log_str(Level.INFO, LogTag.unspecified, [message])


def debug_str(message):  # type: (str) -> LogEntry
    return log_str(Level.DEBUG, LogTag.unspecified, [message])



import datetime

from six.moves.urllib.parse import quote as url_quote


def format_trashinfo(original_location,  # type: str
                     deletion_date,  # type: datetime.datetime
                     ):
    content = ("[Trash Info]\n" +
               "Path=%s\n" % format_original_location(original_location) +
               "DeletionDate=%s\n" % format_date(deletion_date)).encode('utf-8')
    return content


def format_date(deletion_date):  # type: (datetime.datetime) -> str
    return deletion_date.strftime("%Y-%m-%dT%H:%M:%S")


def format_original_location(original_location):  # type: (str) -> str
    return url_quote(original_location, '/')



from typing import NamedTuple, List


class TrashAllResult(NamedTuple('TrashAllResult', [
    ('failed_paths', List[str]),
])):
    def any_failure(self):  # type: () -> bool
        return len(self.failed_paths) > 0



from __future__ import print_function

import six

from trashcli.restore.output import Output
from trashcli.restore.output_event import Println, Die, Quit, Exiting, \
    OutputEvent


class RealOutput(Output):
    def __init__(self, stdout, stderr, exit):
        self.stdout = stdout
        self.stderr = stderr
        self.exit = exit

    def quit(self):
        self.die('')

    def printerr(self, msg):
        print(six.text_type(msg), file=self.stderr)

    def println(self, line):
        print(six.text_type(line), file=self.stdout)

    def die(self, error):
        self.printerr(error)
        self.exit(1)

    def append_event(self,
                     event,  # type: OutputEvent
                     ):
        if isinstance(event, Println):
            self.println(event.msg)
        elif isinstance(event, Die):
            self.die(event.msg)
        elif isinstance(event, Quit):
            self.quit()
        elif isinstance(event, Exiting):
            self.println(event.msg)
        else:
            raise Exception("Unknown call %s" % event)



class Errors:
    def __init__(self, program_name, err):
        self.program_name = program_name
        self.err = err

    def print_error(self, msg):
        self.err.write(format_error_msg(self.program_name, msg))


def format_error_msg(program_name, msg):
    return "%s: %s\n" % (program_name, msg)



# Copyright (C) 2009-2020 Andrea Francia Trivolzio(PV) Italy
import os
from abc import ABCMeta, abstractmethod

import six


@six.add_metaclass(ABCMeta)
class MountPointsListing:
    @abstractmethod
    def list_mount_points(self):
        raise NotImplementedError()


class RealMountPointsListing(MountPointsListing):
    def list_mount_points(self):
        return os_mount_points()


class FakeMountPointsListing(MountPointsListing):
    def __init__(self, mount_points):
        self.mount_points = mount_points

    def set_mount_points(self, mount_points):
        self.mount_points = mount_points

    def list_mount_points(self):
        return self.mount_points


def os_mount_points():
    import psutil
    # List of accepted non-physical fstypes
    fstypes = [
        'nfs',
        'nfs4',
        'p9',  # file system used in WSL 2 (Windows Subsystem for Linux)
        'btrfs',
        'fuse',  # https://github.com/andreafrancia/trash-cli/issues/250
        'fuse.glusterfs',
        # https://github.com/andreafrancia/trash-cli/issues/255
        'fuse.mergerfs',
        'fuse.gocryptfs',
    ]

    # Append fstypes of physical devices to list
    fstypes += set([p.fstype for p in psutil.disk_partitions()])

    partitions = Partitions(fstypes)

    for p in psutil.disk_partitions(all=True):
        if os.path.isdir(p.mountpoint) and \
                partitions.should_used_by_trashcli(p):
            yield p.mountpoint


class Partitions:
    def __init__(self, physical_fstypes):
        self.physical_fstypes = physical_fstypes

    def should_used_by_trashcli(self, partition):
        if ((partition.device, partition.mountpoint,
             partition.fstype) ==
                ('tmpfs', '/tmp', 'tmpfs')):
            return True
        return partition.fstype in self.physical_fstypes



from trashcli.lib.my_input import Input


class User:
    def __init__(self,
                 prepare_output_message,
                 input, # type: Input
                 parse_reply):
        self.prepare_output_message = prepare_output_message
        self.input = input
        self.parse_reply = parse_reply

    def do_you_wanna_empty_trash_dirs(self, trash_dirs):
        reply = self.input.read_input(self.prepare_output_message(trash_dirs))
        return self.parse_reply(reply)



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from __future__ import absolute_import

from trashcli.parse_trashinfo.basket import Basket
from trashcli.parse_trashinfo.parse_trashinfo import ParseTrashInfo


def maybe_parse_deletion_date(contents):
    result = Basket(unknown_date)
    parser = ParseTrashInfo(
        on_deletion_date=lambda date: result.collect(date),
        on_invalid_date=lambda: result.collect(unknown_date)
    )
    parser.parse_trashinfo(contents)
    return result.collected


unknown_date = '????-??-?? ??:??:??'



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from __future__ import absolute_import

import pwd
from typing import Union

from trashcli.lib.trash_dirs import (
    home_trash_dir_path_from_env,
    home_trash_dir_path_from_home)


class UserInfo:
    def __init__(self, home_trash_dir_paths, uid):
        self.home_trash_dir_paths = home_trash_dir_paths
        self.uid = uid


class SingleUserInfoProvider:
    @staticmethod
    def get_user_info(environ, uid):
        return [UserInfo(home_trash_dir_path_from_env(environ), uid)]


class AllUsersInfoProvider:
    @staticmethod
    def get_user_info(_environ, _uid):
        for user in pwd.getpwall():
            yield UserInfo([home_trash_dir_path_from_home(user.pw_dir)],
                           user.pw_uid)


UserInfoProvider = Union[SingleUserInfoProvider, AllUsersInfoProvider]



import os

from trashcli.fstab.volume_of import VolumeOf


class VolumeOfImpl(VolumeOf):
    def __init__(self, ismount, abspath):
        self.ismount = ismount
        self.abspath = abspath

    def volume_of(self, path):
        path = self.abspath(path)
        while path != os.path.dirname(path):
            if self.ismount.is_mount(path):
                break
            path = os.path.dirname(path)
        return path



# Copyright (C) 2011-2021 Andrea Francia Bereguardo(PV) Italy
from trashcli.rm.file_remover import FileRemover
from trashcli.lib.path_of_backup_copy import path_of_backup_copy


class CleanableTrashcan:
    def __init__(self,
                 file_remover, # type: FileRemover
                 ):
        self._file_remover = file_remover

    def delete_trash_info_and_backup_copy(self, trash_info_path):
        backup_copy = path_of_backup_copy(trash_info_path)
        self._file_remover.remove_file_if_exists(backup_copy)
        self._file_remover.remove_file2(trash_info_path)



import os
import posixpath
import re
from typing import NamedTuple

from trashcli.put.core.check_type import CheckType
from trashcli.put.core.path_maker_type import PathMakerType
from trashcli.put.gate import Gate


class Candidate(NamedTuple('Candidate', [
    ('trash_dir_path', str),
    ('volume', str),
    ('path_maker_type', PathMakerType),
    ('check_type', CheckType),
    ('gate', Gate),
])):
    def parent_dir(self):
        return os.path.dirname(self.trash_dir_path)

    def info_dir(self):
        return os.path.join(self.trash_dir_path, 'info')

    def files_dir(self):
        return os.path.join(self.trash_dir_path, 'files')

    def norm_path(self):
        return os.path.normpath(self.trash_dir_path)

    def shrink_user(self, environ):
        path = self.norm_path()
        if environ.get('TRASH_PUT_DISABLE_SHRINK', '') == '1':
            return path
        home_dir = environ.get('HOME', '')
        home_dir = posixpath.normpath(home_dir)
        if home_dir != '':
            path = re.sub('^' + re.escape(home_dir + os.path.sep),
                          '~' + os.path.sep, path)
        return path



from __future__ import absolute_import

import datetime
from six.moves.urllib.parse import unquote


def do_nothing(*argv, **argvk): pass


class ParseTrashInfo:
    def __init__(self,
                 on_deletion_date=do_nothing,
                 on_invalid_date=do_nothing,
                 on_path=do_nothing):
        self.found_deletion_date = on_deletion_date
        self.found_invalid_date = on_invalid_date
        self.found_path = on_path

    def parse_trashinfo(self, contents):
        found_deletion_date = False
        for line in contents.split('\n'):
            if not found_deletion_date and line.startswith('DeletionDate='):
                found_deletion_date = True
                try:
                    date = datetime.datetime.strptime(
                        line, "DeletionDate=%Y-%m-%dT%H:%M:%S")
                except ValueError:
                    self.found_invalid_date()
                else:
                    self.found_deletion_date(date)

            if line.startswith('Path='):
                path = unquote(line[len('Path='):])
                self.found_path(path)



from __future__ import absolute_import

import os

from trashcli.lib.dir_reader import DirReader


class TrashDirReader:

    def __init__(self,
                 dir_reader, # type: DirReader
                 ):
        self.dir_reader = dir_reader

    def list_orphans(self, path):
        info_dir = os.path.join(path, 'info')
        files_dir = os.path.join(path, 'files')
        for entry in self.dir_reader.entries_if_dir_exists(files_dir):
            trashinfo_path = os.path.join(info_dir, entry + '.trashinfo')
            file_path = os.path.join(files_dir, entry)
            if not self.dir_reader.exists(trashinfo_path):
                yield file_path

    def list_trashinfo(self, path):
        info_dir = os.path.join(path, 'info')
        for entry in self.dir_reader.entries_if_dir_exists(info_dir):
            if entry.endswith('.trashinfo'):
                yield os.path.join(info_dir, entry)



import os

from trashcli.put.core.path_maker_type import PathMakerType
from trashcli.put.fs.fs import Fs


class OriginalLocation:
    def __init__(self,
                 fs,  # type: Fs
                 ):
        self.fs = fs

    def for_file(self,
                 path,
                 path_maker_type,  # type: PathMakerType
                 volume_top_dir,
                 ):  # type: (...) -> str
        normalized_path = os.path.normpath(path)
        basename = os.path.basename(normalized_path)
        parent = self.fs.parent_realpath2(normalized_path)
        parent = self._calc_parent_path(parent, volume_top_dir,
                                        path_maker_type)

        return os.path.join(parent, basename)

    @staticmethod
    def _calc_parent_path(parent,  # type: str
                          volume_top_dir,  # type: str
                          path_maker_type,  # type: PathMakerType
                          ):
        if path_maker_type == PathMakerType.AbsolutePaths:
            return parent
        if path_maker_type == PathMakerType.RelativePaths:
            if (parent == volume_top_dir) or parent.startswith(
                    volume_top_dir + os.path.sep):
                parent = parent[len(volume_top_dir + os.path.sep):]
            return parent



from six.moves import range


class Range:
    def __init__(self, start, stop):
        self.start = start
        self.stop = stop

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        if self.start != other.start:
            return False
        if self.stop != other.stop:
            return False
        return True

    def __iter__(self):
        return iter(range(self.start, self.stop + 1))

    def __repr__(self):
        return "Range(%s, %s)" % (self.start, self.stop)



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
import os
import sys

import trashcli.trash
from .file_system import RealRestoreReadFileSystem, \
    RealRestoreWriteFileSystem, RealReadCwd, RealFileReader, \
    RealListingFileSystem
from .info_dir_searcher import InfoDirSearcher
from .info_files import InfoFiles
from .real_restore_logger import RealRestoreLogger
from .restore_cmd import RestoreCmd
from .trash_directories import TrashDirectoriesImpl
from .trashed_files import TrashedFiles
from ..fstab.volumes import RealVolumes
from ..lib.logger import my_logger
from ..lib.my_input import RealInput


def main():
    info_files = InfoFiles(RealListingFileSystem())
    volumes = RealVolumes()
    trash_directories = TrashDirectoriesImpl(volumes,
                                             os.getuid(),
                                             os.environ)
    searcher = InfoDirSearcher(trash_directories, info_files)
    trashed_files = TrashedFiles(RealRestoreLogger(my_logger),
                                 RealFileReader(),
                                 searcher)
    RestoreCmd.make(
        stdout=sys.stdout,
        stderr=sys.stderr,
        exit=sys.exit,
        input=RealInput(),
        version=trashcli.trash.version,
        trashed_files=trashed_files,
        read_fs=RealRestoreReadFileSystem(),
        write_fs=RealRestoreWriteFileSystem(),
        read_cwd=RealReadCwd()
    ).run(sys.argv)



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from __future__ import print_function

from typing import NamedTuple

from trashcli.lib.environ import Environ


class PrintTimeArgs(
    NamedTuple('PrintTimeArgs', [
        ('environ', Environ),
    ])):
    pass


class PrintTimeAction:
    def __init__(self, out, clock):
        self.out = out
        self.clock = clock

    def run_action(self,
                   args,  # type: PrintTimeArgs
                   ):
        now_value = self.clock.get_now_value(args.environ)
        print(now_value.replace(microsecond=0).isoformat(), file=self.out)



import os

from trashcli.restore.file_system import RestoreWriteFileSystem, \
    RestoreReadFileSystem
from trashcli.restore.trashed_file import TrashedFile


class Restorer:
    def __init__(self,
                 read_fs, # type: RestoreReadFileSystem
                 write_fs, # type: RestoreWriteFileSystem
                 ):
        self.read_fs = read_fs
        self.write_fs = write_fs

    def restore_trashed_file(self,
                             trashed_file, # type: TrashedFile
                             overwrite, # type: bool
                             ):
        """
        If overwrite is enabled, then the restore functionality will overwrite an existing file
        """
        if not overwrite and self.read_fs.path_exists(trashed_file.original_location):
            raise IOError(
                'Refusing to overwrite existing file "%s".' % os.path.basename(
                    trashed_file.original_location))
        else:
            parent = os.path.dirname(trashed_file.original_location)
            self.write_fs.mkdirs(parent)

        self.write_fs.move(trashed_file.original_file, trashed_file.original_location)
        self.write_fs.remove_file(trashed_file.info_file)



import os.path
from typing import NamedTuple

from trashcli.put.core.either import Either
from trashcli.put.core.either import Left
from trashcli.put.core.either import Right
from trashcli.put.core.failure_reason import FailureReason
from trashcli.put.core.failure_reason import LogContext
from trashcli.put.fs.fs import Fs
from trashcli.put.janitor_tools.info_file_persister import TrashedFile


class UnableToMoveFileToTrash(NamedTuple('UnableToMoveFileToTrash', [
    ('error', Exception),
]), FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return "failed to move %s in %s: %s" % (
            context.trashee_path,
            context.files_dir(),
            self.error,
        )


class PutTrashDir:
    def __init__(self,
                 fs,  # type: Fs
                 ):
        self.fs = fs

    def try_trash(self,
                  path,  # type: str
                  paths,  # type: TrashedFile
                  ):  # type: (...) -> Either[None, Exception]
        try:
            move_file(self.fs, path, paths.backup_copy_path)
            return Right(None)
        except (IOError, OSError) as error:
            self.fs.remove_file(paths.trashinfo_path)
            return Left(UnableToMoveFileToTrash(error))


def move_file(fs,  # type: Fs
              src, # type: str
              dest, # type: str
              ):
    # Using nornpath allow to delete symlink to a dir even if the are
    # specified with traling slash
    fs.move(os.path.normpath(src), dest)



import os

from typing import NamedTuple

from trashcli.put.clock import PutClock
from trashcli.put.core.candidate import Candidate
from trashcli.put.core.either import Either
from trashcli.put.core.either import Left
from trashcli.put.core.either import Right
from trashcli.put.core.failure_reason import FailureReason
from trashcli.put.core.failure_reason import LogContext
from trashcli.put.format_trash_info import format_trashinfo
from trashcli.put.fs.fs import Fs
from trashcli.put.janitor_tools.info_file_persister import TrashinfoData
from trashcli.put.original_location import OriginalLocation


class UnableToCreateTrashInfoContent(
    NamedTuple('UnableToCreateTrashInfoContent', [
        ('error', Exception),
    ]), FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return "failed to generate trashinfo content: %s" % (
            self.error)


class TrashInfoCreator:
    def __init__(self,
                 fs,   # type: Fs
                 clock,  # type: PutClock
                 ):
        self.original_location = OriginalLocation(fs)
        self.clock = clock

    Result = Either[TrashinfoData, UnableToCreateTrashInfoContent]

    def make_trashinfo_data(self,
                            path,  # type: str
                            candidate,  # type: Candidate
                            ):  # type: (...) -> Result
        try:
            original_location = self.original_location.for_file(path,
                                                                candidate.path_maker_type,
                                                                candidate.volume)
            content = format_trashinfo(original_location, self.clock.now())
            basename = os.path.basename(original_location)
            trash_info_data = TrashinfoData(basename, content,
                                            candidate.info_dir())
            return Right(trash_info_data)
        except (IOError, OSError) as error:
            return Left(UnableToCreateTrashInfoContent(error))



from trashcli.put.core.int_generator import IntGenerator


class Suffix:
    def __init__(self,
                 int_gen,  # type: IntGenerator
                 ):
        self.int_gen = int_gen

    def suffix_for_index(self, index):
        if index == 0:
            return ""
        elif index < 100:
            return "_%s" % index
        else:
            return "_%s" % self.int_gen.new_int(0, 65535)



from __future__ import absolute_import


class Basket:
    def __init__(self, initial_value=None):
        self.collected = initial_value

    def collect(self, value):
        self.collected = value



import os


class Describer:
    def __init__(self, fs):
        self.fs = fs

    def describe(self, path):
        """
        Return a textual description of the file pointed by this path.
        Options:
         - "symbolic link"
         - "directory"
         - "'.' directory"
         - "'..' directory"
         - "regular file"
         - "regular empty file"
         - "non existent"
         - "entry"
        """
        if self.fs.islink(path):
            return 'symbolic link'
        elif self.fs.isdir(path):
            if path == '.':
                return 'directory'
            elif path == '..':
                return 'directory'
            else:
                if os.path.basename(path) == '.':
                    return "'.' directory"
                elif os.path.basename(path) == '..':
                    return "'..' directory"
                else:
                    return 'directory'
        elif self.fs.isfile(path):
            if self.fs.getsize(path) == 0:
                return 'regular empty file'
            else:
                return 'regular file'
        elif not self.fs.exists(path):
            return 'non existent'
        else:
            return 'entry'



import os

from typing import Dict

Environ = Dict[str, str]


def cast_environ(env,
                 ):  # type: (...) -> Environ
    if env == os.environ:
        return env
    else:
        raise ValueError("env must be os.environ")



from typing import NamedTuple, Iterable, Optional

from trashcli.restore.info_files import InfoFiles
from trashcli.restore.trash_directories import TrashDirectories


class InfoDirSearcher:
    def __init__(self,
                 trash_directories,  # type: TrashDirectories
                 info_files, # type: InfoFiles
                 ):
        self.trash_directories = trash_directories
        self.info_files = info_files

    def all_file_in_info_dir(self,
                             trash_dir_from_cli,  # type: Optional[str]
                             ): # type: (...) -> Iterable[FileFound]
        for trash_dir_path, volume in self.trash_directories.list_trash_dirs(
                trash_dir_from_cli):
            for type, path in self.info_files.all_info_files(trash_dir_path):
                yield FileFound(type, path, volume)


class FileFound(NamedTuple('Info', [
    ('type', 'str'),
    ('path', 'str'),
    ('volume', 'str'),
])):
    pass



from typing import NamedTuple

from trashcli.put.core.candidate import Candidate
from trashcli.put.core.either import Either, Right, Left
from trashcli.put.core.failure_reason import FailureReason, LogContext
from trashcli.put.dir_maker import DirMaker
from trashcli.put.fs.fs import Fs


class TrashDirCannotBeCreated(
    NamedTuple('TrashDirCannotBeCreated', [
        ('error', Exception),
    ]), FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return "error during directory creation: %s" % (
            self.error)


class TrashDirCreator:
    def __init__(self, fs):  # type: (Fs) -> None
        self.dir_maker = DirMaker(fs)

    def make_candidate_dirs(self,
                            candidate,  # type: Candidate
                            ):  # type: (...) -> Either[None, TrashDirCannotBeCreated]
        try:
            self.dir_maker.mkdir_p(candidate.trash_dir_path, 0o700)
            self.dir_maker.mkdir_p(candidate.files_dir(), 0o700)
            self.dir_maker.mkdir_p(candidate.info_dir(), 0o700)
            return Right(None)
        except (IOError, OSError) as error:
            return Left(TrashDirCannotBeCreated(error))



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from trashcli.put.fs.fs import Fs
from trashcli.put.context import Context
from trashcli.put.core.trash_result import TrashResult
from trashcli.put.core.trashee import Trashee
from trashcli.put.fs.parent_realpath import ParentRealpathFs
from trashcli.put.fs.volume_of_parent import VolumeOfParent
from trashcli.put.janitor import Janitor
from trashcli.put.my_logger import LoggerBackend
from trashcli.put.my_logger import MyLogger
from trashcli.put.reporting.trash_put_reporter import TrashPutReporter
from trashcli.put.trash_directories_finder import TrashDirectoriesFinder


class FileTrasher:

    def __init__(self,
                 fs,  # type: Fs
                 janitor,  # type: Janitor
                 backend,  # type: LoggerBackend
                 ):  # type: (...) -> None
        self.trash_directories_finder = TrashDirectoriesFinder(fs)
        self.parent_realpath_fs = ParentRealpathFs(fs)
        self.logger = MyLogger(backend)
        self.reporter = TrashPutReporter(fs)
        self.janitor = janitor
        self.volume_of_parent = VolumeOfParent(fs)

    def trash_file(self,
                   path,  # type: str
                   context,  # type: Context
                   ):
        volume = self._figure_out_volume(path, context.forced_volume)
        trashee = Trashee(path, volume)
        candidates = self._select_candidates(volume, context.user_trash_dir,
                                             context.environ,
                                             context.uid, context.home_fallback)
        failures = []
        for candidate in candidates:
            self.logger.log_put(self.reporter.trash_dir_with_volume(candidate),
                                context.log_data)
            trashing = self.janitor.trash_file_in(
                candidate, context.log_data, context.environ, trashee)
            if trashing.succeeded():
                self.logger.log_put(
                    self.reporter.file_has_been_trashed_in_as(path,
                                                              candidate,
                                                              context.environ),
                    context.log_data)
                return TrashResult.Success
            else:
                failures.append((candidate, trashing.reason))
        self.logger.log_put(self.reporter.unable_to_trash_file(
            trashee, failures, context.environ), context.log_data)
        return TrashResult.Failure

    def _figure_out_volume(self, path, default_volume):
        if default_volume:
            return default_volume
        else:
            return self.volume_of_parent.volume_of_parent(path)

    def _select_candidates(self, volume, user_trash_dir, environ, uid,
                           home_fallback):
        return self.trash_directories_finder. \
            possible_trash_directories_for(volume,
                                           user_trash_dir, environ, uid,
                                           home_fallback)



# Copyright (C) 2011-2021 Andrea Francia Bereguardo(PV) Italy
from trashcli.compat import Protocol

from trashcli.fs import ContentsOf
from trashcli.lib.dir_checker import DirChecker
from trashcli.lib.dir_reader import DirReader
from trashcli.lib.user_info import SingleUserInfoProvider
from trashcli.rm.cleanable_trashcan import CleanableTrashcan
from trashcli.rm.file_remover import FileRemover
from trashcli.rm.filter import Filter
from trashcli.rm.list_trashinfo import ListTrashinfos
from trashcli.trash_dirs_scanner import TrashDirsScanner, TopTrashDirRules, \
    trash_dir_found


class RmFileSystemReader(ContentsOf,
                         DirReader,
                         TopTrashDirRules.Reader,
                         Protocol):
    pass


class RmCmd:
    def __init__(self,
                 environ,
                 getuid,
                 volumes_listing,
                 stderr,
                 file_reader,  # type: RmFileSystemReader
                 ):
        self.environ = environ
        self.getuid = getuid
        self.volumes_listing = volumes_listing
        self.stderr = stderr
        self.file_reader = file_reader

    def run(self, argv, uid):
        args = argv[1:]
        self.exit_code = 0

        if not args:
            self.print_err('Usage:\n'
                           '    trash-rm PATTERN\n'
                           '\n'
                           'Please specify PATTERN.\n'
                           'trash-rm uses fnmatch.fnmatchcase to match patterns, see https://docs.python.org/3/library/fnmatch.html for more details.')
            self.exit_code = 8
            return

        trashcan = CleanableTrashcan(FileRemover())
        cmd = Filter(args[0])

        listing = ListTrashinfos.make(self.file_reader, self.file_reader)

        user_info_provider = SingleUserInfoProvider()
        scanner = TrashDirsScanner(user_info_provider,
                                   self.volumes_listing,
                                   TopTrashDirRules(self.file_reader),
                                   DirChecker())

        for event, args in scanner.scan_trash_dirs(self.environ, uid):
            if event == trash_dir_found:
                path, volume = args
                for type, arg in listing.list_from_volume_trashdir(path,
                                                                   volume):
                    if type == 'unable_to_parse_path':
                        self.unable_to_parse_path(arg)
                    elif type == 'trashed_file':
                        original_location, info_file = arg
                        if cmd.matches(original_location):
                            trashcan.delete_trash_info_and_backup_copy(
                                info_file)

    def unable_to_parse_path(self, trashinfo):
        self.report_error('{}: unable to parse \'Path\''.format(trashinfo))

    def report_error(self, error_msg):
        self.print_err('trash-rm: {}'.format(error_msg))

    def print_err(self, msg):
        self.stderr.write(msg + '\n')



from __future__ import print_function

import os
from typing import List
from typing import NamedTuple

from trashcli.lib.dir_reader import DirReader
from trashcli.lib.path_of_backup_copy import path_of_backup_copy
from trashcli.lib.trash_dir_reader import TrashDirReader
from trashcli.list.extractors import DeletionDateExtractor
from trashcli.list.extractors import SizeExtractor
from trashcli.parse_trashinfo.parse_path import parse_path
from trashcli.parse_trashinfo.parser_error import ParseError
from trashcli.trash_dirs_scanner import trash_dir_found
from trashcli.trash_dirs_scanner import \
    trash_dir_skipped_because_parent_is_symlink
from trashcli.trash_dirs_scanner import \
    trash_dir_skipped_because_parent_not_sticky


class ListTrashArgs(
    NamedTuple('ListTrashArgs', [
        ('trash_dirs', List[str]),
        ('attribute_to_print', str),
        ('show_files', bool),
        ('all_users', bool),
    ])):
    pass


class ListTrashAction:
    def __init__(self,
                 environ,
                 uid,
                 selector,
                 out,
                 err,
                 dir_reader,
                 content_reader
                 ):
        self.environ = environ
        self.uid = uid
        self.selector = selector
        self.out = out
        self.err = err
        self.dir_reader = dir_reader
        self.content_reader = content_reader

    def run_action(self,
                   args, # type: ListTrashArgs
                   ):
        for message in ListTrash(self.environ,
                                 self.uid,
                                 self.selector,
                                 self.dir_reader,
                                 self.content_reader).list_all_trash(args):
            self.print_event(message)

    def print_event(self, event):
        if isinstance(event, Error):
            print(event.error, file=self.err)
        elif isinstance(event, Output):
            print(event.message, file=self.out)


class ListTrash:
    def __init__(self,
                 environ,
                 uid,
                 selector,
                 dir_reader,  # type: DirReader
                 content_reader,
                 ):
        self.environ = environ
        self.uid = uid
        self.selector = selector
        self.dir_reader = dir_reader
        self.content_reader = content_reader

    def list_all_trash(self,
                       args,  # type: ListTrashArgs
                       ):
        extractors = {
            'deletion_date': DeletionDateExtractor(),
            'size': SizeExtractor(),
        }
        user_specified_trash_dirs = args.trash_dirs
        extractor = extractors[args.attribute_to_print]
        show_files = args.show_files
        all_users = args.all_users
        trash_dirs = self.selector.select(all_users,
                                          user_specified_trash_dirs,
                                          self.environ,
                                          self.uid)
        for event, event_args in trash_dirs:
            if event == trash_dir_found:
                path, volume = event_args
                trash_dir = TrashDirReader(self.dir_reader)
                for trash_info in trash_dir.list_trashinfo(path):
                    for msg in self._print_trashinfo(volume, trash_info,
                                                     extractor, show_files):
                        yield msg
            elif event == trash_dir_skipped_because_parent_not_sticky:
                path, = event_args
                msg = Error(
                    self.top_trashdir_skipped_because_parent_not_sticky(path))
                yield msg
            elif event == trash_dir_skipped_because_parent_is_symlink:
                path, = event_args
                msg = Error(
                    self.top_trashdir_skipped_because_parent_is_symlink(path))
                yield msg

    def _print_trashinfo(self,
                         volume,
                         trashinfo_path,
                         extractor,
                         show_files):
        try:
            contents = self.content_reader.contents_of(trashinfo_path)
        except IOError as e:
            yield Error(str(e))
        else:
            try:
                relative_location = parse_path(contents)
            except ParseError:
                yield Error(self.print_parse_path_error(trashinfo_path))
            else:
                attribute = extractor.extract_attribute(trashinfo_path,
                                                        contents)
                original_location = os.path.join(volume, relative_location)

                if show_files:
                    original_file = path_of_backup_copy(trashinfo_path)
                    line = format_line2(attribute, original_location,
                                        original_file)
                else:
                    line = format_line(attribute, original_location)
                yield Output(line)

    def top_trashdir_skipped_because_parent_is_symlink(self, trashdir):
        return "TrashDir skipped because parent is symlink: %s" % trashdir

    def top_trashdir_skipped_because_parent_not_sticky(self, trashdir):
        return "TrashDir skipped because parent not sticky: %s" % trashdir

    def print_parse_path_error(self, offending_file):
        return "Parse Error: %s: Unable to parse Path." % offending_file


class Event:
    pass


class Error(Event):
    def __init__(self, error):
        self.error = error


class Output(Event):
    def __init__(self, message):
        self.message = message


def format_line(attribute, original_location):
    return "%s %s" % (attribute, original_location)


def format_line2(attribute, original_location, original_file):
    return "%s %s -> %s" % (attribute, original_location, original_file)



import datetime


class PutClock:
    def now(self):  # type: () -> datetime.datetime
        raise NotImplementedError()


class RealClock(PutClock):
    def now(self):  # type: () -> datetime.datetime
        return datetime.datetime.now()



import os
from abc import ABCMeta, abstractmethod

import six

from trashcli.fstab.mount_points_listing import MountPointsListing, \
    RealMountPointsListing


@six.add_metaclass(ABCMeta)
class VolumesListing:
    @abstractmethod
    def list_volumes(self, environ):  # type (dict) -> Iterable[str]
        raise NotImplementedError()


class FixedVolumesListing(VolumesListing):
    def __init__(self, volumes):
        self.volumes = volumes

    def list_volumes(self, _environ):
        return self.volumes


class RealVolumesListing(VolumesListing):
    def list_volumes(self, environ):
        return VolumesListingImpl(RealMountPointsListing()).list_volumes(
            environ)


class VolumesListingImpl:
    def __init__(self,
                 mount_points_listing,  # type: MountPointsListing
                 ):
        self.mount_points_listing = mount_points_listing

    def list_volumes(self, environ):
        if 'TRASH_VOLUMES' in environ and environ['TRASH_VOLUMES']:
            return [vol
                    for vol in environ['TRASH_VOLUMES'].split(':')
                    if vol != '']
        return self.mount_points_listing.list_mount_points()


class NoVolumesListing(VolumesListing):
    def list_volumes(self, environ):
        return []


class RealIsMount:
    def is_mount(self, path):
        return os.path.ismount(path)



from logging import Logger

from trashcli.restore.restore_logger import RestoreLogger


class RealRestoreLogger(RestoreLogger):
    def __init__(self,
                 logger,  # type: Logger
                 ):
        self._logger = logger

    def warning(self, message):
        self._logger.warning(message)



import os
from abc import abstractmethod
from typing import Iterable

from trashcli.compat import Protocol


class RealPathFs(Protocol):
    @abstractmethod
    def realpath(self, path):
        raise NotImplementedError


class Fs(RealPathFs, Protocol):
    @abstractmethod
    def atomic_write(self, path, content):
        raise NotImplementedError

    @abstractmethod
    def chmod(self, path, mode):
        raise NotImplementedError

    @abstractmethod
    def isdir(self, path):
        raise NotImplementedError

    @abstractmethod
    def isfile(self, path):
        raise NotImplementedError

    @abstractmethod
    def getsize(self, path):
        raise NotImplementedError

    @abstractmethod
    def exists(self, path):
        raise NotImplementedError

    @abstractmethod
    def makedirs(self, path, mode):
        raise NotImplementedError

    @abstractmethod
    def move(self, path, dest):
        raise NotImplementedError

    @abstractmethod
    def remove_file(self, path):
        raise NotImplementedError

    @abstractmethod
    def islink(self, path):
        raise NotImplementedError

    @abstractmethod
    def has_sticky_bit(self, path):
        raise NotImplementedError

    @abstractmethod
    def is_accessible(self, path):
        raise NotImplementedError

    @abstractmethod
    def read(self, path):
        raise NotImplementedError

    @abstractmethod
    def write_file(self, path, content):
        raise NotImplementedError

    @abstractmethod
    def make_file(self, path, content):
        raise NotImplementedError

    @abstractmethod
    def get_mod(self, path):
        raise NotImplementedError

    @abstractmethod
    def lexists(self, path):
        raise NotImplementedError

    @abstractmethod
    def walk_no_follow(self, top):
        raise NotImplementedError

    def parent_realpath2(self, path):
        parent = os.path.dirname(path)
        return self.realpath(parent)

    def list_sorted(self, path):
        return sorted(self.listdir(path))



def list_all(fs, path):  # type: (Fs, str) -> Iterable[str]
    result = fs.walk_no_follow(path)
    for top, dirs, non_dirs in result:
        for d in dirs:
            yield os.path.join(top, d)
        for f in non_dirs:
            yield os.path.join(top, f)



from trashcli.fs import RealExists, RealIsStickyDir, RealIsSymLink
from trashcli.trash_dirs_scanner import TopTrashDirRules


class RealTopTrashDirRulesReader(
    TopTrashDirRules.Reader,
    RealExists,
    RealIsStickyDir,
    RealIsSymLink,
):
    pass



import os
from abc import ABCMeta, abstractmethod
from trashcli.compat import Protocol

import six

from trashcli import fs
from trashcli.fs import FsMethods, RealListFilesInDir, ListFilesInDir, \
    RealContentsOf


class FileReader(Protocol):
    @abstractmethod
    def contents_of(self, path):
        raise NotImplementedError()


class RealFileReader(RealContentsOf, FileReader):
    pass


class FakeFileReader(FileReader):
    def __init__(self, contents=None):
        self.contents = contents

    def set_content(self, contents):
        self.contents = contents

    def contents_of(self, path):
        return self.contents


@six.add_metaclass(ABCMeta)
class RestoreReadFileSystem:
    @abstractmethod
    def path_exists(self, path):  # type: (str) -> bool
        raise NotImplementedError()


class RealRestoreReadFileSystem(RestoreReadFileSystem):
    def path_exists(self, path):
        return os.path.exists(path)


@six.add_metaclass(ABCMeta)
class RestoreWriteFileSystem:
    @abstractmethod
    def mkdirs(self, path):  # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def move(self, path, dest):  # type: (str, str) -> None
        raise NotImplementedError()

    @abstractmethod
    def remove_file(self, path):  # type: (str) -> None
        raise NotImplementedError()


class RealRestoreWriteFileSystem(RestoreWriteFileSystem):
    def mkdirs(self, path):
        return fs.mkdirs(path)

    def move(self, path, dest):
        return fs.move(path, dest)

    def remove_file(self, path):
        return fs.remove_file(path)


@six.add_metaclass(ABCMeta)
class ReadCwd:
    @abstractmethod
    def getcwd_as_realpath(self):  # type: () -> str
        raise NotImplementedError()


class RealReadCwd(ReadCwd):
    def getcwd_as_realpath(self):
        return os.path.realpath(os.curdir)


class FakeReadCwd(ReadCwd):
    def __init__(self, default_cur_dir=None):
        self.default_cur_dir = default_cur_dir

    def chdir(self, path):
        self.default_cur_dir = path

    def getcwd_as_realpath(self):
        return self.default_cur_dir


class ListingFileSystem(ListFilesInDir, Protocol):
    pass


class RealListingFileSystem(ListingFileSystem, RealListFilesInDir):
    pass



from __future__ import absolute_import

from trashcli.compat import Protocol

from trashcli.fs import EntriesIfDirExists, PathExists, RealEntriesIfDirExists, \
    RealExists


class DirReader(
    EntriesIfDirExists,
    PathExists,
    Protocol,
):
    pass


class RealDirReader(RealEntriesIfDirExists, RealExists):
    pass



from typing import NamedTuple

from trashcli.lib.environ import Environ
from trashcli.put.core.candidate import Candidate
from trashcli.put.core.either import Either
from trashcli.put.core.either import Left
from trashcli.put.core.either import Right
from trashcli.put.core.failure_reason import FailureReason
from trashcli.put.core.failure_reason import LogContext
from trashcli.put.core.trashee import Trashee
from trashcli.put.fs.fs import Fs
from trashcli.put.gate import Gate
from trashcli.put.trash_dir_volume_reader import TrashDirVolumeReader


class DifferentVolumes(NamedTuple('DifferentVolumes', [
    ('trash_dir_volume', str),
    ('file_volume', str),
]), FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return (
                "trash dir and file to be trashed are not in the same volume, trash-dir volume: %s, file volume: %s"
                % (self.trash_dir_volume, self.file_volume))


class HomeFallBackNotEnabled(FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return "home fallback not enabled"

    def __eq__(self, other):
        return isinstance(other, HomeFallBackNotEnabled)


GateCheckResult = Either[None, FailureReason]


class TrashDirChecker:
    def __init__(self,
                 fs,  # type: Fs
                 ):  # type: (...) -> None
        self.fs = fs

    def file_could_be_trashed_in(self,
                                 trashee,  # type: Trashee
                                 candidate,  # type: Candidate
                                 environ,  # type: Environ
                                 ):  # type: (...) -> GateCheckResult
        if candidate.gate is Gate.HomeFallback:
            return self._can_be_trashed_in_home_trash_dir(environ)
        elif candidate.gate is Gate.SameVolume:
            return SameVolumeGateImpl(self.fs).can_trash_in(
                trashee, candidate)
        else:
            raise ValueError("Unknown gate: %s" % candidate.gate)

    @staticmethod
    def _can_be_trashed_in_home_trash_dir(environ,  # type: Environ
                                          ):
        if environ.get('TRASH_ENABLE_HOME_FALLBACK', None) == "1":
            return make_ok()
        return Left(HomeFallBackNotEnabled())


def make_ok():
    return Right(None)


class SameVolumeGateImpl:
    def __init__(self,
                 fs,  # type: Fs
                 ):
        self.fs = fs

    def can_trash_in(self,
                     trashee,  # type: Trashee
                     candidate,  # type: Candidate
                     ):
        trash_dir_volume = self._volume_of_trash_dir(candidate)
        same_volume = trash_dir_volume == trashee.volume

        if not same_volume:
            return Left(DifferentVolumes(trash_dir_volume, trashee.volume))

        return make_ok()

    def _volume_of_trash_dir(self, candidate):  # type: (Candidate) -> str
        return (TrashDirVolumeReader(self.fs)
                .volume_of_trash_dir(candidate.trash_dir_path))



from trashcli.put.core.candidate import Candidate
from trashcli.put.core.check_type import NoCheck, TopTrashDirCheck
from trashcli.put.core.either import Either, Right, Left
from trashcli.put.core.failure_reason import FailureReason, LogContext


class SecurityCheck:

    def __init__(self, fs):
        self.fs = fs

    def check_trash_dir_is_secure(self,
                                  candidate,  # type: Candidate
                                  ):  # type: (...) -> Either[None, FailureReason]
        if candidate.check_type == NoCheck:
            return Right(None)
        if candidate.check_type == TopTrashDirCheck:
            parent = candidate.parent_dir()
            if not self.fs.lexists(parent):
                return Left(TrashDirDoesNotHaveParent())
            if not self.fs.isdir(parent):
                return Left(TrashDirCannotBeCreatedBecauseParentIsFile())
            if self.fs.islink(parent):
                return Left(TrashDirIsNotSecureBecauseSymLink())
            if not self.fs.has_sticky_bit(parent):
                return Left(TrashDirIsNotSecureBecauseNotSticky())
            return Right(None)
        raise Exception("Unknown check type: %s" % candidate.check_type)


class TrashDirDoesNotHaveParent(FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return trash_dir_parent_problem(context, (
            "trash dir cannot be created because its parent does not exists"))


class TrashDirCannotBeCreatedBecauseParentIsFile(FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return trash_dir_parent_problem(context, (
            "trash dir cannot be created as its parent is a file instead of being a directory"))


class TrashDirIsNotSecureBecauseSymLink(FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return trash_dir_parent_problem(context, (
            "trash dir is insecure, its parent should not be a symlink"))


class TrashDirIsNotSecureBecauseNotSticky(FailureReason):
    def log_entries(self, context):  # type: (LogContext) -> str
        return trash_dir_parent_problem(context, (
            "trash dir is insecure, its parent should be sticky"))


def trash_dir_parent_problem(context,  # type: LogContext
                             message,  # type: str
                             ):  # type: (...) -> str
    return "%s, trash-dir: %s, parent: %s" % (
        message,
        context.trash_dir_norm_path(),
        context.candidate.parent_dir())



import os

from trashcli.put.fs.fs import RealPathFs


class ParentRealpathFs:
    def __init__(self, fs):  # type: (RealPathFs) -> None
        self.fs = fs

    def parent_realpath(self, path):
        parent = os.path.dirname(path)
        return self.fs.realpath(parent)



import os
from abc import ABCMeta, abstractmethod

import six
from typing import Optional, Iterable

from trashcli.restore.args import RunRestoreArgs
from trashcli.restore.sort_method import sort_files
from trashcli.restore.trashed_file import TrashedFile
from trashcli.restore.trashed_files import TrashedFiles


class RunRestoreAction:
    def __init__(self,
                 handler,  # type: 'Handler'
                 trashed_files,  # type: TrashedFiles
                 ):
        self.handler = handler
        self.trashed_files = trashed_files

    def run_action(self, args,  # type: RunRestoreArgs
                   ):  # type: (...) -> None
        trashed_files = self.all_files_trashed_from_path(args.path,
                                                         args.trash_dir)

        trashed_files = sort_files(args.sort, trashed_files)

        self.handler.handle_trashed_files(trashed_files,
                                          args.overwrite)

    def all_files_trashed_from_path(self,
                                    path,  # type: str
                                    trash_dir_from_cli,  # type: Optional[str]
                                    ):  # type: (...) -> Iterable[TrashedFile]
        for trashed_file in self.trashed_files.all_trashed_files(
                trash_dir_from_cli):
            if trashed_file.original_location_matches_path(path):
                yield trashed_file


@six.add_metaclass(ABCMeta)
class Handler:
    @abstractmethod
    def handle_trashed_files(self,
                             trashed_files,
                             overwrite,  # type: bool
                             ):
        raise NotImplementedError()


def original_location_matches_path(trashed_file_original_location, path):
    if path == os.path.sep:
        return True
    if trashed_file_original_location.startswith(path + os.path.sep):
        return True
    return trashed_file_original_location == path



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from __future__ import absolute_import

import os

# Error codes (from os on *nix, hard coded for Windows):
EX_OK = getattr(os, 'EX_OK', 0)
EX_USAGE = getattr(os, 'EX_USAGE', 64)
EX_IOERR = getattr(os, 'EX_IOERR', 74)



# Copyright (C) 2011-2021 Andrea Francia Bereguardo(PV) Italy
import os
import sys

from trashcli.fs import RealExists, RealIsStickyDir, RealIsSymLink, \
    RealContentsOf, RealEntriesIfDirExists
from trashcli.fstab.volume_listing import RealVolumesListing
from trashcli.rm.rm_cmd import RmCmd, RmFileSystemReader


def main():
    volumes_listing = RealVolumesListing()
    cmd = RmCmd(environ=os.environ,
                getuid=os.getuid,
                volumes_listing=volumes_listing,
                stderr=sys.stderr,
                file_reader=RealRmFileSystemReader())

    cmd.run(sys.argv, os.getuid())

    return cmd.exit_code


class RealRmFileSystemReader(RmFileSystemReader,
                             RealExists,
                             RealIsStickyDir,
                             RealIsSymLink,
                             RealContentsOf,
                             RealEntriesIfDirExists,
                             ):
    pass



from trashcli.trash_dirs_scanner import trash_dir_found


def prepare_output_message(trash_dirs):
    result = []
    if trash_dirs:
        result.append("Would empty the following trash directories:")
        for event, args in trash_dirs:
            if event == trash_dir_found:
                trash_dir, volume = args
                result.append("    - %s" % trash_dir)
        result.append("Proceed? (y/N) ")
        return "\n".join(result)
    else:
        return 'No trash directories to empty.\n'



import os
import shutil
import stat
from abc import abstractmethod

from typing import Iterable, List
from trashcli.compat import Protocol


class FileSize(Protocol):
    @abstractmethod
    def file_size(self, path):
        raise NotImplementedError()


class MakeFileExecutable(Protocol):
    @abstractmethod
    def make_file_executable(self, path):
        raise NotImplementedError()


class WriteFile(Protocol):
    @abstractmethod
    def write_file(self, name, contents):
        raise NotImplementedError()


class ReadFile(Protocol):
    @abstractmethod
    def read_file(self, path):
        raise NotImplementedError()


class Move(Protocol):
    @abstractmethod
    def move(self, path, dest):
        raise NotImplementedError()


class RemoveFile2(Protocol):
    @abstractmethod
    def remove_file2(self, path):
        raise NotImplementedError()


class RemoveFile(Protocol):
    @abstractmethod
    def remove_file(self, path):
        raise NotImplementedError()


class RemoveFileIfExists(Protocol):
    @abstractmethod
    def remove_file_if_exists(self, path):
        raise NotImplementedError()


class EntriesIfDirExists(Protocol):
    @abstractmethod
    def entries_if_dir_exists(self, path):  # type: (str) -> List[str]
        raise NotImplementedError()


class PathExists(Protocol):
    @abstractmethod
    def exists(self, path):
        raise NotImplementedError()


class HasStickyBit(Protocol):
    @abstractmethod
    def has_sticky_bit(self, path):  # type: (str) -> bool
        raise NotImplementedError


class IsStickyDir(Protocol):
    @abstractmethod
    def is_sticky_dir(self, path):  # type: (str) -> bool
        raise NotImplementedError


class IsSymLink(Protocol):
    @abstractmethod
    def is_symlink(self, path):  # type: (str) -> bool
        raise NotImplementedError


class ContentsOf(Protocol):
    @abstractmethod
    def contents_of(self, path):
        raise NotImplementedError()


class RealEntriesIfDirExists(EntriesIfDirExists):
    def entries_if_dir_exists(self, path):
        if os.path.exists(path):
            for entry in os.listdir(path):
                yield entry


class RealExists(PathExists):
    def exists(self, path):  # type: (str) -> bool
        return os.path.exists(path)


class RealHasStickyBit(HasStickyBit):
    def has_sticky_bit(self, path):
        return (os.stat(path).st_mode & stat.S_ISVTX) == stat.S_ISVTX


class RealIsStickyDir(IsStickyDir, RealHasStickyBit):
    def is_sticky_dir(self, path):  # type: (str) -> bool
        return os.path.isdir(path) and self.has_sticky_bit(path)


class RealIsSymLink(IsSymLink):
    def is_symlink(self, path):  # type: (str) -> bool
        return os.path.islink(path)


class RealContentsOf(ContentsOf):
    def contents_of(self, path):
        return _read_file(path)


class RealRemoveFile(RemoveFile):
    def remove_file(self, path):
        if os.path.lexists(path):
            try:
                os.remove(path)
            except:
                return shutil.rmtree(path)


class RealRemoveFile2(RemoveFile2):
    def remove_file2(self, path):
        try:
            os.remove(path)
        except OSError:
            shutil.rmtree(path)


class RealRemoveFileIfExists(RemoveFileIfExists, RemoveFile2):
    def remove_file_if_exists(self, path):
        if os.path.lexists(path): self.remove_file2(path)


class RealMove(Move):
    def move(self, path, dest):
        return shutil.move(path, str(dest))


class ListFilesInDir(Protocol):
    @abstractmethod
    def list_files_in_dir(self, path):  # type: (str) -> Iterable[str]
        raise NotImplementedError()


class RealListFilesInDir(ListFilesInDir):
    def list_files_in_dir(self, path):  # type: (str) -> Iterable[str]
        for entry in os.listdir(path):
            result = os.path.join(path, entry)
            yield result


class MkDirs(Protocol):
    @abstractmethod
    def mkdirs(self, path):
        raise NotImplementedError()


class RealMkDirs(MkDirs):
    def mkdirs(self, path):
        if os.path.isdir(path):
            return
        os.makedirs(path)


class AtomicWrite(Protocol):
    @abstractmethod
    def atomic_write(self, path, content):
        raise NotImplementedError()

    @abstractmethod
    def open_for_write_in_exclusive_and_create_mode(self, path):
        raise NotImplementedError()


class RealAtomicWrite(AtomicWrite):
    def atomic_write(self, path, content):
        file_handle = self.open_for_write_in_exclusive_and_create_mode(path)
        os.write(file_handle, content)
        os.close(file_handle)

    def open_for_write_in_exclusive_and_create_mode(self, path):
        return os.open(path, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0o600)


class RealReadFile(ReadFile):
    def read_file(self, path):
        return _read_file(path)


class RealWriteFile(WriteFile):
    def write_file(self, name, contents):
        with open(name, 'w') as f:
            f.write(contents)


class RealMakeFileExecutable(MakeFileExecutable):
    def make_file_executable(self, path):
        os.chmod(path, os.stat(path).st_mode | stat.S_IXUSR)


class RealFileSize(FileSize):
    def file_size(self, path):
        return os.stat(path).st_size


class FsMethods(
    RealEntriesIfDirExists,
    RealExists,
    RealIsStickyDir,
    RealIsSymLink,
    RealContentsOf,
    RealRemoveFile,
    RealRemoveFile2,
    RealRemoveFileIfExists,
    RealMove,
    RealListFilesInDir,
    RealMkDirs,
    RealAtomicWrite,
    RealReadFile,
    RealWriteFile,
    RealMakeFileExecutable,
    RealFileSize,
):
    pass


def _read_file(path):
    with open(path) as f:
        return f.read()


has_sticky_bit = RealHasStickyBit().has_sticky_bit
contents_of = RealContentsOf().contents_of
remove_file = RealRemoveFile().remove_file
move = RealMove().move
list_files_in_dir = RealListFilesInDir().list_files_in_dir
mkdirs = RealMkDirs().mkdirs
atomic_write = RealAtomicWrite().atomic_write
open_for_write_in_exclusive_and_create_mode = RealAtomicWrite().open_for_write_in_exclusive_and_create_mode
read_file = RealReadFile().read_file
write_file = RealWriteFile().write_file
make_file_executable = RealMakeFileExecutable().make_file_executable
file_size = RealFileSize().file_size
remove_file2 = RealRemoveFile2().remove_file2
is_sticky_dir = RealIsStickyDir().is_sticky_dir



# Copyright (C) 2011-2021 Andrea Francia Bereguardo(PV) Italy
import os
from abc import abstractmethod
from trashcli.compat import Protocol

from trashcli.fs import ContentsOf
from trashcli.lib.dir_reader import DirReader
from trashcli.lib.trash_dir_reader import TrashDirReader
from trashcli.parse_trashinfo.parse_path import parse_path
from trashcli.parse_trashinfo.parser_error import ParseError


class FileContentReader(Protocol):
    @abstractmethod
    def contents_of(self, path):
        raise NotImplementedError()


class ListTrashinfos:
    def __init__(self,
                 file_content_reader,  # type: ContentsOf
                 trash_dir_reader,  # type: TrashDirReader
                 ):
        self.trash_dir_reader = trash_dir_reader
        self.file_content_reader = file_content_reader

    def list_from_volume_trashdir(self, trashdir_path, volume):
        for trashinfo_path in self.trash_dir_reader.list_trashinfo(
                trashdir_path):
            trashinfo = self.file_content_reader.contents_of(trashinfo_path)
            try:
                path = parse_path(trashinfo)
            except ParseError:
                yield 'unable_to_parse_path', trashinfo_path
            else:
                complete_path = os.path.join(volume, path)
                yield 'trashed_file', (complete_path, trashinfo_path)

    @staticmethod
    def make(file_content_reader,  # type: ContentsOf
             dir_reader,  # type: DirReader
             ):
        trash_dir_reader = TrashDirReader(dir_reader)
        return ListTrashinfos(file_content_reader, trash_dir_reader)



from typing import List

from trashcli.restore.output import Output
from trashcli.restore.output_event import Quit, Die, Println, OutputEvent


class OutputRecorder(Output):
    def __init__(self):  # type: (...) -> None
        self.events = []  # type: List[OutputEvent]

    def quit(self):  # type: () -> None
        self.append_event(Quit())

    def die(self, msg):  # type: (str) -> None
        self.append_event(Die(msg))

    def println(self, msg):  # type: (str) -> None
        self.append_event(Println(msg))

    def append_event(self,
                     event,  # type: OutputEvent
                     ):  # type: (...) -> None
        self.events.append(event)

    def apply_to(self, output,  # type: Output
                 ):   # type: (...) -> None
        for event in self.events:
            output.append_event(event)



from typing import List, NamedTuple

from trashcli.trash_dirs_scanner import trash_dir_found, \
    trash_dir_skipped_because_parent_not_sticky, \
    trash_dir_skipped_because_parent_is_symlink


class ListTrashDirsArgs(
    NamedTuple('ListTrashDirsArgs',
               [('trash_dirs', List[str]),
                ('all_users', bool)])
):
    pass


class ListTrashDirs:
    def __init__(self, environ, uid, selector):
        self.environ = environ
        self.uid = uid
        self.selector = selector

    def run_action(self, args):
        user_specified_trash_dirs = args.trash_dirs
        all_users = args.all_users
        trash_dirs = self.selector.select(all_users,
                                          user_specified_trash_dirs,
                                          self.environ,
                                          self.uid)
        for event, event_args in trash_dirs:
            if event == trash_dir_found:
                path, volume = event_args
                print("%s" % path)
            elif event == trash_dir_skipped_because_parent_not_sticky:
                path = event_args
                print("parent_not_sticky: %s" % (path))
            elif event == trash_dir_skipped_because_parent_is_symlink:
                path = event_args
                print("parent_is_symlink: %s" % (path))



import errno
import os
from typing import NamedTuple, Iterator

from trashcli.lib.path_of_backup_copy import path_of_backup_copy
from trashcli.put.fs.fs import Fs
from trashcli.put.jobs import JobStatus, NeedsMoreAttempts, Succeeded, \
    JobExecutor
from trashcli.put.my_logger import LogData, MyLogger
from trashcli.put.suffix import Suffix


class TrashinfoData(NamedTuple('TrashinfoData', [
    ('basename', str),
    ('content', str),
    ('info_dir_path', str),
])):
    pass


class TrashedFile(NamedTuple('TrashedFile', [
    ('trashinfo_path', str),
])):
    @property
    def backup_copy_path(self):  # type: () -> str
        return path_of_backup_copy(self.trashinfo_path)


class InfoFilePersister:
    def __init__(self,
                 fs,  # type: Fs
                 logger,  # type: MyLogger
                 suffix,  # type: Suffix
                 ):  # type: (...) -> None
        self.fs = fs
        self.logger = logger
        self.suffix = suffix

    def create_trashinfo_file(self,
                              trashinfo_data,  # type: TrashinfoData
                              log_data,  # type: LogData
                              ):  # type: (...) -> TrashedFile
        return JobExecutor(self.logger, TrashedFile).execute(
            self.try_persist(trashinfo_data), log_data)

    Result = Iterator[JobStatus[TrashedFile]]

    def try_persist(self,
                    data,  # type: TrashinfoData
                    ):  # type: (...) -> Result
        index = 0
        name_too_long = False
        while True:
            suffix = self.suffix.suffix_for_index(index)
            trashinfo_basename = create_trashinfo_basename(data.basename,
                                                           suffix,
                                                           name_too_long)
            trashinfo_path = os.path.join(data.info_dir_path,
                                          trashinfo_basename)
            if os.path.exists(path_of_backup_copy(trashinfo_path)):
                index += 1
                continue
            try:
                self.fs.atomic_write(trashinfo_path, data.content)
                yield Succeeded(TrashedFile(trashinfo_path),
                                ".trashinfo created as %s." % trashinfo_path)
            except OSError as e:
                if e.errno == errno.ENAMETOOLONG:
                    name_too_long = True
                yield NeedsMoreAttempts(trashinfo_path,
                                        "attempt for creating %s failed." % trashinfo_path)

            index += 1


def create_trashinfo_basename(basename, suffix, name_too_long):
    after_basename = suffix + ".trashinfo"
    if name_too_long:
        truncated_basename = basename[0:len(basename) - len(after_basename)]
    else:
        truncated_basename = basename
    return truncated_basename + after_basename



# Copyright (C) 2007-2023 Andrea Francia Trivolzio(PV) Italy
from typing import List, Dict, Iterator, Tuple

from trashcli.fstab.volume_of import VolumeOf
from trashcli.lib.dir_checker import DirChecker
from trashcli.lib.user_info import AllUsersInfoProvider, \
    SingleUserInfoProvider
from trashcli.trash_dirs_scanner import trash_dir_found, TrashDir, \
    TopTrashDirRules, TrashDirsScanner


class TrashDirsSelector:
    def __init__(self,
                 current_user_dirs,
                 all_users_dirs,
                 volumes  # type: VolumeOf
                 ):
        self.current_user_dirs = current_user_dirs
        self.all_users_dirs = all_users_dirs
        self.volumes = volumes

    def select(self,
               all_users_flag, # type: bool
               user_specified_dirs, # type: List[str]
               environ, # type: Dict[str, str]
               uid, # type: int
               ):  # type: (...) -> Iterator[Tuple[str, TrashDir]]
        if all_users_flag:
            for dir in self.all_users_dirs.scan_trash_dirs(environ, uid):
                yield dir
        else:
            if not user_specified_dirs:
                for dir in self.current_user_dirs.scan_trash_dirs(environ, uid):
                    yield dir
            for dir in user_specified_dirs:
                yield trash_dir_found, (
                    TrashDir(dir, self.volumes.volume_of(dir)))

    @staticmethod
    def make(volumes_listing,
             reader,  # type: TopTrashDirRules.Reader
             volumes  # type: VolumeOf
             ):
        user_info_provider = SingleUserInfoProvider()
        user_dir_scanner = TrashDirsScanner(user_info_provider,
                                            volumes_listing,
                                            TopTrashDirRules(reader),
                                            DirChecker())
        all_users_info_provider = AllUsersInfoProvider()
        all_users_scanner = TrashDirsScanner(all_users_info_provider,
                                             volumes_listing,
                                             TopTrashDirRules(reader),
                                             DirChecker())
        return TrashDirsSelector(user_dir_scanner,
                                 all_users_scanner,
                                 volumes)



from __future__ import absolute_import

from trashcli.parse_trashinfo.parse_trashinfo import ParseTrashInfo
from trashcli.parse_trashinfo.basket import Basket


def parse_deletion_date(contents):
    result = Basket()
    parser = ParseTrashInfo(on_deletion_date=result.collect)
    parser.parse_trashinfo(contents)
    return result.collected



# Copyright (C) 2011-2021 Andrea Francia Bereguardo(PV) Italy
import fnmatch
import os


class Filter:
    def __init__(self, pattern):
        self.pattern = pattern

    def matches(self, original_location):
        basename = os.path.basename(original_location)
        subject = original_location if self.pattern[0] == '/' else basename
        return fnmatch.fnmatchcase(subject, self.pattern)



import argparse
from enum import Enum
from typing import List, Union

from trashcli.lib.print_version import PrintVersionArgs
from trashcli.list.list_trash_action import ListTrashArgs
from trashcli.list.minor_actions.debug_volumes import DebugVolumesArgs
from trashcli.list.minor_actions.list_trash_dirs import ListTrashDirsArgs
from trashcli.list.minor_actions.list_volumes import PrintVolumesArgs
from trashcli.list.minor_actions.print_python_executable import \
    PrintPythonExecutableArgs
from trashcli.shell_completion import add_argument_to, TRASH_DIRS

Args = Union[
    PrintVersionArgs,
    PrintVolumesArgs,
    DebugVolumesArgs,
    ListTrashDirsArgs,
    ListTrashArgs,
    PrintPythonExecutableArgs
]


class Parser:
    def __init__(self, prog):
        self.parser = argparse.ArgumentParser(prog=prog,
                                              description='List trashed files',
                                              epilog='Report bugs to https://github.com/andreafrancia/trash-cli/issues')
        add_argument_to(self.parser)
        self.parser.add_argument('--version',
                                 dest='action',
                                 action='store_const',
                                 const=ListAction.print_version,
                                 default=ListAction.list_trash,
                                 help="show program's version number and exit")
        self.parser.add_argument('--debug-volumes',
                                 dest='action',
                                 action='store_const',
                                 const=ListAction.debug_volumes,
                                 help=argparse.SUPPRESS)
        self.parser.add_argument('--volumes',
                                 dest='action',
                                 action='store_const',
                                 const=ListAction.list_volumes,
                                 help="list volumes")
        self.parser.add_argument('--trash-dirs',
                                 dest='action',
                                 action='store_const',
                                 const=ListAction.list_trash_dirs,
                                 help="list trash dirs")
        self.parser.add_argument('--trash-dir',
                                 action='append',
                                 default=[],
                                 dest='trash_dirs',
                                 help='specify the trash directory to use'
                                 ).complete = TRASH_DIRS
        self.parser.add_argument('--size',
                                 action='store_const',
                                 default='deletion_date',
                                 const='size',
                                 dest='attribute_to_print',
                                 help=argparse.SUPPRESS)
        self.parser.add_argument('--files',
                                 action='store_true',
                                 default=False,
                                 dest='show_files',
                                 help=argparse.SUPPRESS)
        self.parser.add_argument('--all-users',
                                 action='store_true',
                                 dest='all_users',
                                 help='list trashcans of all the users')
        self.parser.add_argument('--python',
                                 dest='action',
                                 action='store_const',
                                 const=ListAction.print_python_executable,
                                 help=argparse.SUPPRESS)

    def parse_list_args(self,
                        args,  # type: List[str]
                        argv0,  # type: str
                        ):  # type: (...) -> Args

        parsed = self.parser.parse_args(args)

        if parsed.action == ListAction.print_version:
            return PrintVersionArgs(argv0=argv0)
        if parsed.action == ListAction.list_volumes:
            return PrintVolumesArgs()
        if parsed.action == ListAction.debug_volumes:
            return DebugVolumesArgs()
        if parsed.action == ListAction.list_trash_dirs:
            return ListTrashDirsArgs(
                trash_dirs=parsed.trash_dirs,
                all_users=parsed.all_users
            )
        if parsed.action == ListAction.list_trash:
            return ListTrashArgs(
                trash_dirs=parsed.trash_dirs,
                attribute_to_print=parsed.attribute_to_print,
                show_files=parsed.show_files,
                all_users=parsed.all_users
            )
        if parsed.action == ListAction.print_python_executable:
            return PrintPythonExecutableArgs()

        raise ValueError('Unknown action: {}'.format(parsed.action))


class ListAction(Enum):
    debug_volumes = 'debug_volumes'
    print_version = 'print_version'
    list_trash = 'list_trash'
    list_volumes = 'list_volumes'
    list_trash_dirs = 'list_trash_dirs'
    print_python_executable = 'print_python_executable'



from enum import Enum


class Mode(Enum):
    mode_unspecified = 'mode_unspecified'
    mode_interactive = 'mode_interactive'
    mode_force = 'mode_force'

    def can_ignore_not_existent_path(self):  # type: () -> bool
        return self == Mode.mode_force

    def should_we_ask_to_the_user(self, is_path_accessible):  # type: (bool) -> bool
        return self == Mode.mode_interactive and is_path_accessible



import grp
import os
import pwd
import stat
from typing import NamedTuple
from typing import Optional

from trashcli import fs
from trashcli.fs import write_file
from trashcli.fstab.real_volume_of import RealVolumeOf
from trashcli.put.fs.fs import Fs


class Names:
    def username(self, uid):  # type: (int) -> Optional[str]
        try:
            return pwd.getpwuid(uid).pw_name
        except KeyError as e:
            return None

    def groupname(self, gid):
        try:
            return grp.getgrgid(gid).gr_name
        except KeyError as e:
            return None


class Stat(NamedTuple('Stat', [
    ('mode', int),
    ('uid', int),
    ('gid', int),
])):
    pass


class RealFs(RealVolumeOf, Fs):

    def __init__(self):
        super(RealFs, self).__init__()

    def readlink(self, path):
        return os.readlink(path)

    def symlink(self, src, dest):  # type: (str, str) -> None
        os.symlink(src, dest)

    def touch(self, path):  # type: (str) -> None
        with open(path, 'a'):
            import os
            os.utime(path, None)

    def atomic_write(self, path, content):
        fs.atomic_write(path, content)

    def chmod(self, path, mode):
        os.chmod(path, mode)

    def isdir(self, path):
        return os.path.isdir(path)

    def isfile(self, path):
        return os.path.isfile(path)

    def getsize(self, path):
        return os.path.getsize(path)

    def walk_no_follow(self, path):
        try:
            import scandir  # type: ignore
            walk = scandir.walk
        except ImportError:
            walk = os.walk

        return walk(path, followlinks=False)

    def exists(self, path):
        return os.path.exists(path)

    def makedirs(self, path, mode):
        os.makedirs(path, mode)

    def lstat(self, path):
        stat = os.lstat(path)
        return Stat(mode=stat.st_mode, uid=stat.st_uid, gid=stat.st_gid)

    def mkdir(self, path):
        os.mkdir(path)

    def mkdir_with_mode(self, path, mode):
        os.mkdir(path, mode)

    def move(self, path, dest):
        return fs.move(path, dest)

    def remove_file(self, path):
        fs.remove_file(path)

    def islink(self, path):
        return os.path.islink(path)

    def has_sticky_bit(self, path):
        return (os.stat(path).st_mode & stat.S_ISVTX) == stat.S_ISVTX

    def realpath(self, path):
        return os.path.realpath(path)

    def is_accessible(self, path):
        return os.access(path, os.F_OK)

    def make_file(self, path, content):
        write_file(path, content)

    def get_mod(self, path):
        return stat.S_IMODE(os.lstat(path).st_mode)

    def listdir(self, path):
        return os.listdir(path)

    def read(self, path):
        return fs.read_file(path)

    def write_file(self, path, content):
        return fs.write_file(path, content)

    def lexists(selfs, path):
        return os.path.lexists(path)



import os
import random
import sys

from trashcli.lib.environ import cast_environ
from trashcli.lib.my_input import Input
from trashcli.lib.my_input import RealInput
from trashcli.put.clock import RealClock
from trashcli.put.core.int_generator import IntGenerator
from trashcli.put.describer import Describer
from trashcli.put.file_trasher import FileTrasher
from trashcli.put.fs.fs import Fs
from trashcli.put.fs.real_fs import RealFs
from trashcli.put.janitor import Janitor
from trashcli.put.my_logger import LoggerBackend
from trashcli.put.my_logger import StreamBackend
from trashcli.put.reporting.trash_put_reporter import TrashPutReporter
from trashcli.put.trash_put_cmd import TrashPutCmd
from trashcli.put.trasher import Trasher
from trashcli.put.user import User


def main():
    cmd = make_cmd(clock=RealClock(),
                   fs=RealFs(),
                   user_input=RealInput(),
                   randint=RandomIntGenerator(),
                   backend=StreamBackend(sys.stderr))
    try:
        uid = int(os.environ["TRASH_PUT_FAKE_UID_FOR_TESTING"])
    except KeyError:
        uid = os.getuid()
    return cmd.run_put(sys.argv, cast_environ(os.environ), uid)


def make_cmd(clock,
             fs,  # type: Fs
             user_input,  # type: Input
             randint,  # type: IntGenerator
             backend,  # type: LoggerBackend
             ):  # type: (...) -> TrashPutCmd
    reporter = TrashPutReporter(fs)
    janitor = Janitor(fs, clock, backend, randint)
    file_trasher = FileTrasher(fs, janitor, backend)
    user = User(user_input, Describer(fs))
    trasher = Trasher(file_trasher, user, fs, backend)
    return TrashPutCmd(reporter, trasher)


class RandomIntGenerator(IntGenerator):
    def new_int(self,
                min,  # type: int
                max,  # type: int
                ):
        return random.randint(min, max)



from typing import List
from typing import NamedTuple
from typing import Optional

from trashcli.compat import Protocol
from trashcli.lib.environ import Environ
from trashcli.put.core.logs import LogData
from trashcli.put.core.mode import Mode
from trashcli.put.core.trash_all_result import TrashAllResult
from trashcli.put.core.trash_result import TrashResult


class Context(NamedTuple('Context', [
    ('paths', List[str]),
    ('user_trash_dir', Optional[str]),
    ('mode', Mode),
    ('forced_volume', Optional[str]),
    ('home_fallback', bool),
    ('program_name', str),
    ('log_data', LogData),
    ('environ', Environ),
    ('uid', int),
])):
    def trash_each(self, trasher,  # type: SingleTrasher
                   ):  # type (...) -> TrashAllResult
        failed_paths = []
        for path in self.paths:
            result = trasher.trash_single(path, self)
            if result == TrashResult.Failure:
                failed_paths.append(path)

        return TrashAllResult(failed_paths)


class SingleTrasher(Protocol):
    def trash_single(self,
                     path,  # type: str
                     context,  # type: 'Context'
                     ):
        raise NotImplementedError



from typing import Iterable, NamedTuple

from trashcli.empty.user import User
from trashcli.trash_dirs_scanner import TrashDir

UserIntention = NamedTuple('UserIntention',
                           [('ok_to_empty', bool),
                            ('trash_dirs', Iterable[TrashDir])])


class Guard:

    def __init__(self, user):  # type: (User) -> None
        self.user = user

    def ask_the_user(self,
                     parsed_interactive,  # type: bool
                     trash_dirs,  # type: Iterable[TrashDir]
                     ):  # type: (...) -> UserIntention
        if parsed_interactive:
            return self._interactive(trash_dirs)
        else:
            return self.non_interactive(trash_dirs)

    def _interactive(self, trash_dirs,  # type: Iterable[TrashDir]
                     ):  # type:  (...) -> UserIntention
        trash_dirs_list = list(trash_dirs)  # type: Iterable[TrashDir]
        ok_to_empty = \
            self.user.do_you_wanna_empty_trash_dirs(trash_dirs_list)
        list_result = trash_dirs_list if ok_to_empty else []
        return UserIntention(ok_to_empty=ok_to_empty,
                             trash_dirs=list_result)

    def non_interactive(self,
                        trash_dirs,  # type: Iterable[TrashDir]
                        ):
        trash_dirs_list = trash_dirs  # type: Iterable[TrashDir]
        return UserIntention(ok_to_empty=True,
                             trash_dirs=trash_dirs_list)



from trashcli.compat import Protocol


class IntGenerator(Protocol):
    def new_int(self,
                 min,  # type: int
                 max,  # type: int
                 ):  # type: (...) -> int
        raise NotImplementedError



from abc import ABCMeta

import six
import os
from trashcli.fstab.mount_points_listing import MountPointsListing, \
    RealMountPointsListing
from trashcli.fstab.volume_of import VolumeOf
from trashcli.fstab.real_volume_of import RealVolumeOf


@six.add_metaclass(ABCMeta)
class Volumes(VolumeOf, MountPointsListing):
    pass


class RealVolumes(Volumes):
    def volume_of(self, path):
        return RealVolumeOf().volume_of(path)

    def list_mount_points(self):
        return RealMountPointsListing().list_mount_points()


class VolumesImpl(Volumes):
    def __init__(self,
                 volumes,  # type: VolumeOf
                 mount_point_listing,  # type: MountPointsListing
                 ):
        self.volumes = volumes
        self.mount_point_listing = mount_point_listing

    def volume_of(self, path):
        return self.volumes.volume_of(path)

    def list_mount_points(self):
        return self.mount_point_listing.list_mount_points()


class FakeVolumes(Volumes):
    def __init__(self,
                 mount_points,  # type Iterable[str]
                 ):
        self.mount_points = mount_points

    def list_mount_points(self):
        return self.mount_points

    def volume_of(self, path):
        while path != os.path.dirname(path):
            if self.is_a_mount_point(path):
                break
            path = os.path.dirname(path)
        return path

    def is_a_mount_point(self, path):
        return path in self.mount_points

    def add_volume(self, path):
        self.mount_points.append(path)


class FakeVolumes2(Volumes):
    def __init__(self, volume_of_string, volumes_list):
        self.volume_of_string = volume_of_string
        self.volumes_list = volumes_list

    def volume_of(self, path):
        return self.volume_of_string % path

    def set_volumes(self, volumes_list):
        self.volumes_list = volumes_list

    def list_mount_points(self):
        return self.volumes_list



from trashcli.put.fs.fs import Fs


class DirMaker:
    def __init__(self, fs): # type: (Fs) -> None
        self.fs = fs

    def mkdir_p(self, path, mode):
        try:
            self.fs.makedirs(path, mode)
        except OSError:
            if not self.fs.isdir(path):
                raise



from abc import abstractmethod
from typing import Callable, TypeVar, Generic

S = TypeVar("S")
R = TypeVar("R")
E = TypeVar("E")


class Either(Generic[S, E]):
    @abstractmethod
    def bind(self,
             func):  # type: (Callable[[S], Either[R, E]]) -> Either[R, E]
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, other):  # type: (object) -> bool
        raise NotImplementedError

    def is_error(self):  # type: () -> bool
        return not self.is_valid()

    def is_valid(self):  # type: () -> bool
        return {Left: False, Right: True}[type(self)]

    def error(self):  # type: () -> E
        if isinstance(self, Left):
            return self._error
        else:
            raise ValueError("Not an error: %s" % self)

    def value(self):
        if isinstance(self, Right):
            return self._value
        else:
            raise ValueError("Not a value: %s" % self)


class Right(Either[S, E]):
    def __init__(self, value):  # type: (S) -> None
        self._value = value

    def bind(self,
             func):  # type: (Callable[[S], Either[R, E]]) -> Either[R, E]
        return func(self._value)

    def __eq__(self, other):  # type: (object) -> bool
        return isinstance(other, Right) and self._value == other._value

    def __str__(self):  # type: () -> str
        return "Right %s" % self._value


class Left(Either[S, E]):
    def __init__(self, error):  # type: (E) -> None
        self._error = error

    def bind(self,
             func):  # type: (Callable[[S], Either[R, E]]) -> Either[R, E]
        return Left(self._error)

    def __eq__(self, other):  # type: (object) -> bool
        return isinstance(other, Left) and self._error == other._error

    def __str__(self):  # type: () -> str
        return "Left: %s" % self._error



import argparse
from copy import copy

from typing import Dict

try:
    def convert_to_list(tuple):
        return [item for item in tuple]


    from shtab import add_argument_to, FILE, DIR  # type: ignore

    defaults = convert_to_list(add_argument_to.__defaults__)
    defaults[-1] = {
        "zsh": r"""
# https://github.com/zsh-users/zsh/blob/19390a1ba8dc983b0a1379058e90cd51ce156815/Completion/Unix/Command/_rm#L72-L74
_trash_files() {
  (( CURRENT > 0 )) && line[CURRENT]=()
  line=( ${line//(#m)[\[\]()\\*?#<>~\^\|]/\\$MATCH} )
  _files -F line
}
""",
    }
    add_argument_to.__defaults__ = tuple(defaults)
    TRASH_FILES = copy(FILE)
    TRASH_DIRS = copy(DIR)


    def complete_with(completion,  # type: Dict[str, str]
                      action,  # type: argparse.Action
                      ):
        action.complete = completion  # type: ignore



except ImportError:
    from argparse import Action

    TRASH_FILES = TRASH_DIRS = {}


    class PrintCompletionAction(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            print('Please install shtab firstly!')
            parser.exit(0)


    def add_argument_to(parser, *args, **kwargs):  # type: ignore
        Action.complete = None  # type: ignore
        parser.add_argument(
            '--print-completion',
            choices=['bash', 'zsh', 'tcsh'],
            action=PrintCompletionAction,
            help='print shell completion script',
        )
        return parser


    def complete_with(completion,  # type: Dict[str, str]
                      action,  # type: argparse.Action
                      ):
        pass

TRASH_FILES.update({"zsh": "_trash_files"})
TRASH_DIRS.update({"zsh": "(\\${\\$(trash-list --trash-dirs)#parent_*})"})
